import csv
import importlib.util
import json
import sys
from abc import abstractmethod
from collections.abc import Collection, Iterator, Mapping, Sequence
from dataclasses import dataclass
from datetime import date, datetime, timezone
from functools import lru_cache
from io import TextIOWrapper
from pathlib import Path
from types import MappingProxyType
from typing import IO, TYPE_CHECKING, Any, ClassVar, Generic, Literal, SupportsIndex, TypeAlias, TypeVar, overload

from cognite.client.data_classes.data_modeling import data_types as dt
from cognite.client.data_classes.data_modeling.views import MappedProperty, ViewProperty

from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingDependencyError, ToolkitTypeError, ToolkitValueError
from cognite_toolkit._cdf_tk.utils import humanize_collection, sanitize_filename
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump

from .useful_types import JsonVal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import pyarrow as pa
    import pyarrow.parquet as pq

FileFormat: TypeAlias = Literal["csv", "parquet", "yaml"]
DataType: TypeAlias = Literal["string", "integer", "float", "boolean", "json", "date", "timestamp", "epoch"]
PrimaryCellValue: TypeAlias = datetime | date | str | int | float | bool | JsonVal | None
CellValue: TypeAlias = PrimaryCellValue | list[PrimaryCellValue]
Rows: TypeAlias = list[dict[str, CellValue]]


@dataclass(frozen=True)
class SchemaColumn:
    name: str
    type: DataType
    is_array: bool = False

    def __post_init__(self) -> None:
        if self.type == "json" and self.is_array:
            raise ValueError("JSON columns cannot be arrays. Use 'is_array=False' for JSON columns.")


class SchemaColumnList(list, Sequence[SchemaColumn]):
    # Implemented to get correct type hints
    def __init__(self, collection: Collection[SchemaColumn] | None = None) -> None:
        super().__init__(collection or [])

    def __iter__(self) -> Iterator[SchemaColumn]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> SchemaColumn: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> SchemaColumn | Self:
        if isinstance(index, slice):
            return type(self)(super().__getitem__(index))
        return super().__getitem__(index)

    @classmethod
    def create_from_view_properties(cls, properties: Mapping[str, ViewProperty], support_edges: bool = False) -> Self:
        """Create a SchemaColumnList from a mapping of ViewProperty objects.

        Args:
            properties (Mapping[str, ViewProperty]): A mapping of property names to ViewProperty objects.
            support_edges (bool): Whether the the view supports edges. If True, the schema will include
                startNode and endNode columns.

        Returns:
            SchemaColumnList: A list of SchemaColumn objects representing the properties.
        """
        columns = [
            SchemaColumn("space", "string", is_array=False),
            SchemaColumn("externalId", "string", is_array=False),
            SchemaColumn("instanceType", "string"),
            SchemaColumn("existingVersion", "integer", is_array=False),
            SchemaColumn("type", "json", is_array=False),
        ]
        if support_edges:
            columns.append(SchemaColumn("startNode", "json", is_array=False))
            columns.append(SchemaColumn("endNode", "json", is_array=False))
        for name, prop in properties.items():
            if not isinstance(prop, MappedProperty):
                # We skip all properties that does not reside in a container.
                continue
            schema_type = cls._dms_to_schema_type(prop.type)
            is_array = (
                isinstance(prop.type, dt.ListablePropertyType)
                and prop.type.is_list
                and schema_type != "json"  # JSON is not an array type
            )
            columns.append(SchemaColumn(name=f"properties.{name}", type=schema_type, is_array=is_array))
        return cls(columns)

    @classmethod
    def _dms_to_schema_type(cls, model_type: dt.PropertyType) -> DataType:
        if isinstance(model_type, dt.Text | dt.Enum | dt.CDFExternalIdReference):
            return "string"
        elif isinstance(model_type, dt.Boolean):
            return "boolean"
        elif isinstance(model_type, dt.Json | dt.DirectRelation):
            return "json"
        elif isinstance(model_type, dt.Int32 | dt.Int64):
            return "integer"
        elif isinstance(model_type, dt.Float32 | dt.Float64):
            return "float"
        elif isinstance(model_type, dt.Timestamp):
            return "timestamp"
        elif isinstance(model_type, dt.Date):
            return "date"
        else:
            raise ToolkitTypeError(
                f"Failed convertion from data modeling type to Table Schema. Unknown type: {type(model_type)!r}."
            )


@dataclass
class Schema:
    display_name: str
    folder_name: str
    kind: str
    format_: FileFormat
    columns: SchemaColumnList


T_IO = TypeVar("T_IO", bound=IO)


class TableFileWriter(Generic[T_IO]):
    encoding = "utf-8"
    newline = "\n"
    format: ClassVar[FileFormat]

    def __init__(self, schema: Schema, output_dir: Path, max_file_size_bytes: int = 128 * 1024 * 1024) -> None:
        self.max_file_size_bytes = max_file_size_bytes
        self.schema = schema
        self.output_dir = output_dir
        self._file_count = 1
        self._writer_by_filepath: dict[Path, T_IO] = {}

    def write_rows(self, rows_group_list: list[tuple[str, Rows]]) -> None:
        """Write rows to a file."""
        for group, group_rows in rows_group_list:
            if not group_rows:
                continue
            writer = self._get_writer(group)
            self._write_rows(writer, group_rows)

    @abstractmethod
    def _write_rows(self, writer: T_IO, rows: Rows) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _create_writer(self, filepath: Path) -> T_IO:
        """Create a writer for the given file path."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    @abstractmethod
    def _is_above_file_size_limit(self, filepath: Path, writer: T_IO) -> bool:
        """Check if the file size is above the limit."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def __enter__(self) -> "TableFileWriter":
        self._file_count = 1
        return self

    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any | None) -> None:
        for writer in self._writer_by_filepath.values():
            writer.close()
        self._writer_by_filepath.clear()
        return None

    def _get_writer(self, group: str) -> T_IO:
        clean_name = f"{sanitize_filename(group)}-" if group else ""
        file_path = (
            self.output_dir
            / self.schema.folder_name
            / f"{clean_name}part-{self._file_count:04}.{self.schema.kind}.{self.format}"
        )
        file_path.parent.mkdir(parents=True, exist_ok=True)
        if file_path not in self._writer_by_filepath:
            self._writer_by_filepath[file_path] = self._create_writer(file_path)
        elif self._is_above_file_size_limit(file_path, self._writer_by_filepath[file_path]):
            self._writer_by_filepath[file_path].close()
            del self._writer_by_filepath[file_path]
            self._file_count += 1
            return self._get_writer(group)
        return self._writer_by_filepath[file_path]

    @classmethod
    def get_write_cls(cls, format_: FileFormat) -> "type[TableFileWriter]":
        """Get the writer class for the given format."""
        write_cls = _TABLEWRITER_CLASS_BY_FORMAT.get(format_)
        if write_cls is None:
            raise ToolkitValueError(
                f"Unsupported format {format_}. Supported formats are {humanize_collection(_TABLEWRITER_CLASS_BY_FORMAT.keys())}."
            )
        return write_cls


class ParquetWriter(TableFileWriter["pq.ParquetWriter"]):
    """Parquet writer for CDF Toolkit.

    Caveat: This mutates the rows to convert JSON, timestamp, and date columns to appropriate formats.
    This is necessary because pyarrow does not support JSON, timestamp, and date types directly in the way we need.
    We avoid making a copy of each row for performance reasons, but this means that the rows passed to this writer
    will be modified in place.
    """

    format = "parquet"

    def __init__(self, schema: Schema, output_dir: Path, max_file_size_bytes: int = 128 * 1024 * 1024) -> None:
        super().__init__(schema, output_dir, max_file_size_bytes)
        self._check_pyarrow_dependency()

    def _create_writer(self, filepath: Path) -> "pq.ParquetWriter":
        import pyarrow.parquet as pq

        schema = self._create_schema()
        return pq.ParquetWriter(filepath, schema)

    def _write_rows(self, writer: "pq.ParquetWriter", rows: Rows) -> None:
        import pyarrow as pa

        if json_columns := self._json_columns():
            for row in rows:
                json_values = set(row.keys()) & json_columns
                for col in json_values:
                    row[col] = json.dumps(row[col])
        if timestamp_columns := self._timestamp_columns():
            for row in rows:
                for col in set(row.keys()) & timestamp_columns:
                    cell_value = row[col]
                    if isinstance(cell_value, list):
                        # MyPy does not understand that a list of PrimaryCellValue is valid here
                        # It expects a union of PrimaryCellValue and list[PrimaryCellValue].
                        row[col] = [self._to_datetime(value) for value in cell_value]  # type: ignore[assignment]
                    else:
                        row[col] = self._to_datetime(cell_value)
        if date_columns := self._date_columns():
            for row in rows:
                for col in set(row.keys()) & date_columns:
                    cell_value = row[col]
                    if isinstance(cell_value, list):
                        # MyPy does not understand that a list of PrimaryCellValue is valid here.
                        # It expects a union of PrimaryCellValue and list[PrimaryCellValue].
                        row[col] = [self._to_date(value) for value in cell_value]  # type: ignore[assignment]
                    else:
                        row[col] = self._to_date(cell_value)

        table = pa.Table.from_pylist(rows, schema=self._create_schema())
        writer.write_table(table)

    def _is_above_file_size_limit(self, filepath: Path, writer: "pq.ParquetWriter") -> bool:
        return filepath.exists() and filepath.stat().st_size > self.max_file_size_bytes

    @lru_cache(maxsize=1)
    def _json_columns(self) -> set[str]:
        """Check if the writer supports JSON format."""
        return {col.name for col in self.schema.columns if col.type == "json"}

    @lru_cache(maxsize=1)
    def _timestamp_columns(self) -> set[str]:
        """Check if the writer supports timestamp format."""
        return {col.name for col in self.schema.columns if col.type == "timestamp"}

    @lru_cache(maxsize=1)
    def _date_columns(self) -> set[str]:
        return {col.name for col in self.schema.columns if col.type == "date"}

    @classmethod
    def _to_datetime(cls, value: CellValue) -> CellValue:
        if isinstance(value, datetime) or value is None:
            output = value
        elif isinstance(value, date):
            output = datetime.combine(value, datetime.min.time())
        elif isinstance(value, int | float):
            # Assuming the value is a timestamp in milliseconds
            output = datetime.fromtimestamp(value / 1000.0)
        elif isinstance(value, str):
            output = cls._convert_data_modelling_timestamp(value)
        else:
            raise ToolkitTypeError(
                f"Unsupported value type for datetime conversion: {type(value)}. Expected datetime, date, int, float, or str."
            )
        if output is not None and output.tzinfo is None:
            # Ensure the datetime is in UTC
            output = output.replace(tzinfo=timezone.utc)
        elif output is not None and output.tzinfo is not None:
            # Convert to UTC if it has a timezone
            output = output.astimezone(timezone.utc)
        return output

    @classmethod
    def _to_date(cls, value: CellValue) -> CellValue:
        if isinstance(value, date) or value is None:
            return value
        elif isinstance(value, datetime):
            return value.date()
        elif isinstance(value, int | float):
            # Assuming the value is a timestamp in milliseconds
            return date.fromtimestamp(value / 1000.0)
        elif isinstance(value, str):
            return cls._convert_data_modelling_timestamp(value).date()
        else:
            raise ToolkitTypeError(
                f"Unsupported value type for date conversion: {type(value)}. Expected date, datetime, int, float, or str."
            )

    @classmethod
    def _convert_data_modelling_timestamp(cls, timestamp: str) -> datetime:
        """Convert a timestamp string from the data modeling format to a datetime object."""
        try:
            return datetime.fromisoformat(timestamp)
        except ValueError:
            # Typically hits if the timestamp has truncated milliseconds,
            # For example, "2021-01-01T00:00:00.17+00:00".
            # In Python 3.10, the strptime requires exact formats so we need both formats below.
            # In Python 3.11-13, if the timestamp matches on the second it will match on the first,
            # so when we set lower bound to 3.11 the loop will not be needed.
            for format_ in ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z"]:
                try:
                    return datetime.strptime(timestamp, format_)
                except ValueError:
                    continue
            raise ValueError(
                f"Invalid timestamp format: {timestamp}. Expected ISO 8601 format with optional milliseconds and timezone."
            )

    @lru_cache(maxsize=1)
    def _create_schema(self) -> "pa.Schema":
        """Create a pyarrow schema from the schema definition."""
        self._check_pyarrow_dependency()
        import pyarrow as pa

        fields: list[pa.Field] = []
        for prop in self.schema.columns:
            pa_type = self._as_pa_type(prop.type, prop.is_array)
            fields.append(pa.field(prop.name, pa_type, nullable=True))
        return pa.schema(fields)

    @staticmethod
    def _check_pyarrow_dependency() -> None:
        if importlib.util.find_spec("pyarrow") is None:
            raise ToolkitMissingDependencyError(
                "Writing to parquet requires pyarrow. Install with 'pip install \"cognite-toolkit[table]\"'"
            )

    @staticmethod
    def _as_pa_type(type_: DataType, is_array: bool) -> "pa.DataType":
        """Convert a data type to a pyarrow type."""
        import pyarrow as pa

        if type_ == "string":
            pa_type = pa.string()
        elif type_ == "integer":
            pa_type = pa.int64()
        elif type_ == "float":
            pa_type = pa.float64()
        elif type_ == "boolean":
            pa_type = pa.bool_()
        elif type_ == "date":
            pa_type = pa.date32()
        elif type_ == "time":
            pa_type = pa.time64("ms")
        elif type_ == "json":
            pa_type = pa.string()
        elif type_ == "timestamp":
            pa_type = pa.timestamp("ms", tz="UTC")
        else:
            raise ToolkitValueError(f"Unsupported data type {type_}.")

        if is_array:
            pa_type = pa.list_(pa_type)
        return pa_type


class CSVWriter(TableFileWriter[TextIOWrapper]):
    format = "csv"

    def _create_writer(self, filepath: Path) -> TextIOWrapper:
        stream = filepath.open("a", encoding=self.encoding, newline=self.newline)
        writer = self._create_dict_writer(stream)
        if filepath.stat().st_size == 0:
            writer.writeheader()
        return stream

    def _is_above_file_size_limit(self, filepath: Path, writer: TextIOWrapper) -> bool:
        current_position = writer.tell()
        writer.seek(0, 2)
        if writer.tell() > self.max_file_size_bytes:
            return True
        writer.seek(current_position)
        return False

    def _write_rows(self, writer: TextIOWrapper, rows: Rows) -> None:
        dict_writer = self._create_dict_writer(writer)
        dict_writer.writerows(rows)

    def _create_dict_writer(self, writer: TextIOWrapper) -> csv.DictWriter:
        return csv.DictWriter(writer, fieldnames=[col.name for col in self.schema.columns], extrasaction="ignore")


class YAMLWriter(TableFileWriter[TextIOWrapper]):
    format = "yaml"

    def _create_writer(self, filepath: Path) -> TextIOWrapper:
        return filepath.open("a", encoding=self.encoding, newline=self.newline)

    def _is_above_file_size_limit(self, filepath: Path, writer: TextIOWrapper) -> bool:
        current_position = writer.tell()
        writer.seek(0, 2)
        if writer.tell() > self.max_file_size_bytes:
            return True
        writer.seek(current_position)
        return False

    def _write_rows(self, writer: TextIOWrapper, rows: Rows) -> None:
        writer.write(yaml_safe_dump(rows))


_TABLEWRITER_CLASS_BY_FORMAT: MappingProxyType[str, type[TableFileWriter]] = MappingProxyType(
    {w.format: w for w in TableFileWriter.__subclasses__()}  # type: ignore[type-abstract]
)



from collections.abc import Sequence

import pytest
import responses
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.extended_filemetadata import (
    ExtendedFileMetadataList,
)
from cognite_toolkit._cdf_tk.client.data_classes.pending_instances_ids import PendingInstanceId


class TestExtendedFileMetadataAPI:
    @pytest.mark.parametrize(
        "instance_id, id, external_id, expected_list",
        [
            pytest.param(("my_space", "myExternalId"), None, "my_file", False, id="External ID with tuple"),
            pytest.param(("my_space", "myExternalId"), 123, None, False, id="ID with tuple"),
            pytest.param(NodeId("my_space", "myExternalId"), None, "my_file", False, id="External ID with NodeId"),
            pytest.param(NodeId("my_space", "myExternalId"), 123, None, False, id="ID with NodeId"),
            pytest.param(
                [PendingInstanceId(NodeId("my_space", "myExternalId"), 123)], None, None, True, id="List of pending IDs"
            ),
        ],
    )
    def test_set_pending_instance_ids_valid_inputs(
        self,
        instance_id: NodeId | tuple[str, str] | Sequence[PendingInstanceId],
        id: int | None,
        external_id: str | None,
        expected_list: bool,
        toolkit_config: ToolkitClientConfig,
    ) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/files/set-pending-instance-ids"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"externalId": "does-not-matter", "id": 123, "createdTime": 0, "lastUpdatedTime": 0}]},
            )
            result = client.files.set_pending_ids(instance_id=instance_id, id=id, external_id=external_id)

        is_list = isinstance(result, ExtendedFileMetadataList)
        assert is_list == expected_list, f"Expected result to be a list: {expected_list}, got {is_list}"

    def test_set_instance_ids_invalid(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        with pytest.raises(TypeError) as excinfo:
            client.files.set_pending_ids(("my_space", "MyExternalId", 1), id=123)

        assert (
            str(excinfo.value)
            == "instance_id must be a NodeId, a tuple of (str, str), or a sequence of PendingIdentifier objects."
        )

    @pytest.mark.parametrize(
        "id, external_id, expected_list",
        [
            pytest.param(None, "my_file", False, id="External ID only"),
            pytest.param(123, None, False, id="ID only"),
            pytest.param([123, 456], None, True, id="List of IDs"),
            pytest.param(None, ["my_file", "my_other_file"], True, id="List of External IDs"),
            pytest.param([123, 456], ["my_file", "my_other_file"], True, id="List of IDs and External IDs"),
        ],
    )
    def test_unlink_instance_ids_valid(
        self,
        id: int | Sequence[int] | None,
        external_id: str | SequenceNotStr[str] | None,
        expected_list: bool,
        toolkit_config: ToolkitClientConfig,
    ) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/files/unlink-instance-ids"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"externalId": "does-not-matter", "id": 123, "createdTime": 0, "lastUpdatedTime": 0}]},
            )
            result = client.files.unlink_instance_ids(id=id, external_id=external_id)

        is_list = isinstance(result, ExtendedFileMetadataList)
        assert is_list == expected_list, f"Expected result to be a list: {expected_list}, got {is_list}"

    def test_unlink_instance_ids_no_identifiers_raises(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        with pytest.raises(ValueError, match=r"At least one of id or external_id must be provided."):
            client.files.unlink_instance_ids(id=None, external_id=None)

    def test_unlink_instance_ids_invalid(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        with pytest.raises(
            ValueError, match=r"Cannot specify both id and external_id as single values. Use one or the other."
        ):
            client.files.unlink_instance_ids(id=123, external_id="123")



import importlib.util
import itertools
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable, Mapping
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from functools import cached_property, partial
from itertools import zip_longest
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Generic, Literal, TypeAlias, TypeVar, overload
from zipfile import BadZipFile

import questionary
from cognite.client.data_classes import Transformation
from cognite.client.exceptions import CogniteException
from rich import box
from rich.console import Console
from rich.live import Live
from rich.spinner import Spinner
from rich.table import Table

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawProfileResults, RawTable
from cognite_toolkit._cdf_tk.constants import MAX_ROW_ITERATION_RUN_QUERY
from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingDependencyError, ToolkitThrottledError, ToolkitValueError
from cognite_toolkit._cdf_tk.utils.aggregators import (
    AssetAggregator,
    AssetCentricAggregator,
    EventAggregator,
    FileAggregator,
    LabelAggregator,
    LabelCountAggregator,
    MetadataAggregator,
    RelationshipAggregator,
    SequenceAggregator,
    TimeSeriesAggregator,
)
from cognite_toolkit._cdf_tk.utils.cdf import get_transformation_sources, raw_row_count
from cognite_toolkit._cdf_tk.utils.interactive_select import AssetCentricDestinationSelect, AssetInteractiveSelect
from cognite_toolkit._cdf_tk.utils.sql_parser import SQLParser, SQLTable
from cognite_toolkit._cdf_tk.utils.text import sanitize_spreadsheet_title
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricDestinationType

from ._base import ToolkitCommand

if TYPE_CHECKING:
    from openpyxl.worksheet.worksheet import Worksheet


class WaitingAPICallClass:
    def __bool__(self) -> bool:
        return False


WaitingAPICall = WaitingAPICallClass()


PendingCellValue: TypeAlias = int | float | str | bool | None | WaitingAPICallClass
CellValue: TypeAlias = int | float | str | bool | None

T_Index = TypeVar("T_Index", bound=Hashable)


class ProfileCommand(ToolkitCommand, ABC, Generic[T_Index]):
    spreadsheet_max_column_width = 70.0

    def __init__(
        self,
        output_spreadsheet: Path | None = None,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
    ) -> None:
        super().__init__(print_warning, skip_tracking, silent)
        self.table_title = self.__class__.__name__.removesuffix("Command")
        self.output_spreadsheet: Path | None = output_spreadsheet
        if output_spreadsheet is not None:
            self._validate_openpyxl_installed()

    class Columns:  # Placeholder for columns, subclasses should define their own Columns class
        ...

    spinner_args: ClassVar[Mapping] = dict(name="arc", text="loading...", style="bold green", speed=1.0)

    max_workers = 8
    is_dynamic_table = False

    @cached_property
    def columns(self) -> tuple[str, ...]:
        return (
            tuple([value for attr, value in self.Columns.__dict__.items() if not attr.startswith("_")])
            if hasattr(self, "Columns")
            else tuple()
        )

    @staticmethod
    def _validate_openpyxl_installed() -> None:
        """Ensure that openpyxl is installed if output_spreadsheet is set."""
        if importlib.util.find_spec("openpyxl") is None:
            raise ToolkitMissingDependencyError(
                "Writing to a spreadsheet requires 'openpyxl'. Install with 'pip install \"cognite-toolkit[table]\"'"
            )

    def create_profile_table(self, client: ToolkitClient, sheet: str | None = None) -> list[dict[str, CellValue]]:
        console = Console()
        with console.status("Setting up", spinner="aesthetic", speed=0.4) as _:
            table = self.create_initial_table(client)
        with (
            Live(self.draw_table(table), refresh_per_second=4, console=console) as live,
            ThreadPoolExecutor(max_workers=self.max_workers) as executor,
        ):
            while True:
                current_calls = {
                    #
                    executor.submit(self.create_api_callable(row, col, client)): (row, col)
                    for (row, col), cell in table.items()
                    if cell is WaitingAPICall
                }
                if not current_calls:
                    break
                for future in as_completed(current_calls):
                    row, col = current_calls[future]
                    try:
                        result = future.result()
                    except CogniteException as e:
                        result = type(e).__name__
                    except ToolkitThrottledError as e:
                        result = f"Throttled: Wait {e.wait_time_seconds:.0f} seconds"
                    table[(row, col)] = self.format_result(result, row, col)
                    if self.is_dynamic_table:
                        table = self.update_table(table, result, row, col)
                    live.update(self.draw_table(table))
        result = self.as_record_format(table, allow_waiting_api_call=False)
        if self.output_spreadsheet is not None:
            self._write_to_spreadsheet(result, list(self.columns), self.output_spreadsheet, sheet=sheet)
        return result

    @abstractmethod
    def create_initial_table(self, client: ToolkitClient) -> dict[tuple[T_Index, str], PendingCellValue]:
        """
        Create the initial table with placeholders for API calls.
        Each cell that requires an API call should be initialized with WaitingAPICall.
        """
        raise NotImplementedError("Subclasses must implement create_initial_table.")

    @abstractmethod
    def create_api_callable(self, row: T_Index, col: str, client: ToolkitClient) -> Callable:
        raise NotImplementedError("Subclasses must implement call_api.")

    def format_result(self, result: object, row: T_Index, col: str) -> CellValue:
        """
        Format the result of an API call for display in the table.
        This can be overridden by subclasses to customize formatting.
        """
        if isinstance(result, int | float | bool | str):
            return result
        raise NotImplementedError("Subclasses must implement format_result.")

    def update_table(
        self,
        current_table: dict[tuple[T_Index, str], PendingCellValue],
        result: object,
        row: T_Index,
        col: str,
    ) -> dict[tuple[T_Index, str], PendingCellValue]:
        raise NotImplementedError("Subclasses must implement update_table.")

    def draw_table(self, table: dict[tuple[T_Index, str], PendingCellValue]) -> Table:
        rich_table = Table(
            title=self.table_title,
            title_justify="left",
            show_header=True,
            header_style="bold magenta",
            box=box.MINIMAL,
        )
        for col in self.columns:
            rich_table.add_column(col)

        rows = self.as_record_format(table)

        last_row: list[str | Spinner] = []
        for row in rows:
            this_row = [self._as_cell(value) for value in row.values()]
            draw_row = self._create_draw_row(this_row, last_row)
            rich_table.add_row(*draw_row)
            last_row = this_row
        return rich_table

    @classmethod
    def _create_draw_row(cls, this_row: list[str | Spinner], last_row: list[str | Spinner]) -> list[str | Spinner]:
        """Creates the row to be drawn. This skips sequential cells that have not changed
        such that the table does not have too many repeated values and thus becomes easier to read.
        """
        draw_row: list[str | Spinner] = []
        row_has_changed = False
        for cell, last_cell in zip_longest(this_row, last_row, fillvalue=""):
            if not row_has_changed and cls._should_skip_drawing(cell, last_cell):
                cell = ""
            else:
                # If the first cell in the row has changed, all remaining cells in the row should be drawn.
                row_has_changed = True
            draw_row.append(cell)
        return draw_row

    @classmethod
    def _should_skip_drawing(cls, cell: str | Spinner, last_cell: str | Spinner) -> bool:
        return cell == last_cell or (isinstance(cell, Spinner) and isinstance(last_cell, Spinner))

    @classmethod
    @overload
    def as_record_format(
        cls, table: dict[tuple[T_Index, str], PendingCellValue], allow_waiting_api_call: Literal[True] = True
    ) -> list[dict[str, PendingCellValue]]: ...

    @classmethod
    @overload
    def as_record_format(
        cls,
        table: dict[tuple[T_Index, str], PendingCellValue],
        allow_waiting_api_call: Literal[False],
    ) -> list[dict[str, CellValue]]: ...

    @classmethod
    def as_record_format(
        cls,
        table: dict[tuple[T_Index, str], PendingCellValue],
        allow_waiting_api_call: bool = True,
    ) -> list[dict[str, PendingCellValue]] | list[dict[str, CellValue]]:
        rows: list[dict[str, PendingCellValue]] = []
        row_indices: dict[T_Index, int] = {}
        for (row, col), value in table.items():
            if value is WaitingAPICall and not allow_waiting_api_call:
                value = None
            if row not in row_indices:
                row_indices[row] = len(rows)
                rows.append({col: value})
            else:
                rows[row_indices[row]][col] = value
        return rows

    def _as_cell(self, value: PendingCellValue) -> str | Spinner:
        if isinstance(value, WaitingAPICallClass):
            return Spinner(**self.spinner_args)
        elif isinstance(value, int):
            return f"{value:,}"
        elif isinstance(value, float):
            return f"{value:.2f}"
        elif value is None:
            return "-"
        return str(value)

    def _write_to_spreadsheet(
        self, data: list[dict[str, CellValue]], columns: list[str], output_spreadsheet: Path, sheet: str | None = None
    ) -> None:
        """Write the profile data to a spreadsheet."""
        # Local import as this is an optional dependency
        from openpyxl import Workbook, load_workbook

        sheet_name = sanitize_spreadsheet_title(sheet or self.table_title)[
            :31
        ]  # Limit title to 31 characters for Excel compatibility
        if output_spreadsheet.exists():
            try:
                workbook = load_workbook(output_spreadsheet)
            except (OSError, BadZipFile) as e:
                raise ToolkitValueError(
                    f"Failed to open {output_spreadsheet.as_posix()!r}. "
                    "Please ensure the file is not corrupted or open in another application."
                ) from e
            if sheet_name in workbook.sheetnames:
                raise ToolkitValueError(f"Sheet '{sheet_name}' already exists in {output_spreadsheet.as_posix()}.")
            worksheet = workbook.create_sheet(title=sheet_name)
        else:
            workbook = Workbook()
            worksheet = workbook.active
            worksheet.title = sheet_name

        worksheet.append(columns)

        for row in data:
            worksheet.append(list(row.values()))

        self._style_sheet(worksheet, columns)

        try:
            workbook.save(output_spreadsheet)
        except OSError as e:
            raise ToolkitValueError(
                f"Failed to write to {output_spreadsheet.as_posix()!r}. "
                "Please ensure the file is not open, and that you have sufficient permissions and disk space."
            ) from e
        self.console(f"Profile data written to sheet {sheet!r} in {output_spreadsheet.as_posix()!r}")

    def _style_sheet(self, sheet: "Worksheet", columns: list[str]) -> None:
        """Styles the sheet with the given headers.

        Args:
            sheet: The sheet to style.
            headers: The headers to style.
        """
        # Local import as this is an optional dependency
        from openpyxl.cell import MergedCell
        from openpyxl.styles import Font, PatternFill

        # This freezes all rows above the given row
        sheet.freeze_panes = "A2"

        # Make the header row bold, larger, and colored
        for cell, *_ in sheet.iter_cols(min_row=1, max_row=1, min_col=1, max_col=len(columns)):
            cell.font = Font(bold=True, size=15)
            cell.fill = PatternFill(fgColor="A9DFBF", patternType="solid")
        # Adjust columns width based on widest cell in each column
        for column_cells in sheet.columns:
            try:
                max_length = max(len(str(cell.value)) for cell in column_cells if cell.value is not None)
            except ValueError:
                max_length = 0

            # Find first non-merged cell to safely get column letter
            safe_cell = next((c for c in column_cells if not isinstance(c, MergedCell)), None)
            if safe_cell is None:
                continue  # Skip if no such cell is found

            column_letter = safe_cell.column_letter
            current = sheet.column_dimensions[column_letter].width or (max_length + 0.5)
            sheet.column_dimensions[column_letter].width = min(
                max(current, max_length + 0.5), self.spreadsheet_max_column_width
            )

    def _ask_store_file(self) -> None:
        if file_path := questionary.path("Where do you want to save the profile?").ask():
            self.output_spreadsheet = Path(file_path)


@dataclass(frozen=True)
class AssetIndex:
    aggregator: str
    data_set_external_id: str | None = None
    source: RawTable | None = None


class ProfileAssetCommand(ProfileCommand[AssetIndex]):
    def __init__(
        self,
        output_spreadsheet: Path | None = None,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
    ) -> None:
        super().__init__(output_spreadsheet, print_warning, skip_tracking, silent)
        self.table_title = "Asset Profile for Hierarchy"
        self.hierarchy: str | None = None
        self.aggregators: dict[str, MetadataAggregator] = {}
        self.profile_row_limit = self.max_profile_row_limit

    class Columns:
        Resource = "Resource"
        Count = "Count"
        DataSets = "DataSet"
        DataSetCount = "DataSet Count"
        Transformations = "Transformation"
        RawTable = "Raw Table"
        RowCount = "Rows"
        ColumnCount = "Columns"

    is_dynamic_table = True
    max_profile_row_limit = 10_000  # The number of rows to profile to get the number of columns.
    # The actual limit is 256 MB of data.
    # Ref https://github.com/cognitedata/profiler-api/blob/main/src/main/scala/com/cognite/raw_profiler/Profile.scala#L37
    profile_timeout_seconds = 60 * 4  # Timeout for the profiling operation in seconds,

    def assets(
        self,
        client: ToolkitClient,
        hierarchy: str | None = None,
        profile_row_limit: int = max_profile_row_limit,
        verbose: bool = False,
    ) -> list[dict[str, CellValue]]:
        """
        Profile assets in the given hierarchy.
        This method will create a table with the count of assets, events, files, timeseries, sequences,
        relationships, and labels in the specified hierarchy.
        """
        if hierarchy is None:
            self.hierarchy = AssetInteractiveSelect(client, "profile").select_hierarchy(allow_empty=False)
            self._ask_store_file()
        else:
            self.hierarchy = hierarchy
        if profile_row_limit <= 0 or profile_row_limit > self.max_profile_row_limit:
            raise ToolkitValueError(
                f"Profile row limit must be between 1 and {self.max_profile_row_limit}, got {profile_row_limit}."
            )
        self.table_title = f"Asset Profile for Hierarchy: {self.hierarchy}"
        self.profile_row_limit = profile_row_limit
        self.aggregators = {
            agg.display_name: agg
            for agg in [
                AssetAggregator(client),
                EventAggregator(client),
                FileAggregator(client),
                TimeSeriesAggregator(client),
                SequenceAggregator(client),
            ]
        }
        return self.create_profile_table(client, sheet=hierarchy)

    def create_initial_table(self, client: ToolkitClient) -> dict[tuple[AssetIndex, str], PendingCellValue]:
        table: dict[tuple[AssetIndex, str], PendingCellValue] = {}
        for index, aggregator in self.aggregators.items():
            asset_index = AssetIndex(aggregator=index, data_set_external_id=None, source=None)
            table[(asset_index, self.Columns.Resource)] = aggregator.display_name
            table[(asset_index, self.Columns.Count)] = WaitingAPICall
            table[(asset_index, self.Columns.DataSets)] = None
            table[(asset_index, self.Columns.DataSetCount)] = None
            table[(asset_index, self.Columns.Transformations)] = None
            table[(asset_index, self.Columns.RawTable)] = None
            table[(asset_index, self.Columns.RowCount)] = None
            table[(asset_index, self.Columns.ColumnCount)] = None
        return table

    def create_api_callable(self, row: AssetIndex, col: str, client: ToolkitClient) -> Callable:
        aggregator = self.aggregators[row.aggregator]
        if col == self.Columns.Count:
            return partial(aggregator.count, hierarchy=self.hierarchy)
        elif col == self.Columns.DataSets:
            return partial(aggregator.used_data_sets, hierarchy=self.hierarchy)
        elif col == self.Columns.DataSetCount:
            if row.data_set_external_id is None:
                raise ValueError(f"DataSet external ID is required for {row!s} in column {col}.")
            return partial(aggregator.count, data_set_external_id=row.data_set_external_id, hierarchy=self.hierarchy)
        elif col == self.Columns.Transformations:
            if row.data_set_external_id is None:
                raise ValueError(f"DataSet external ID is required for {row!s} in column {col}.")
            return partial(aggregator.used_transformations, data_set_external_ids=[row.data_set_external_id])
        elif col == self.Columns.ColumnCount:
            if row.source is None:
                raise ValueError(f"Database and table name are required for {row!s} in column {col}.")
            source = row.source
            return partial(
                client.raw.profile,
                database=source.db_name,
                table=source.table_name,
                limit=self.profile_row_limit,
                timeout_seconds=self.profile_timeout_seconds,
            )
        elif col == self.Columns.RowCount:
            if row.source is None:
                raise ValueError(f"Database and table name are required for {row!s} in column {col}.")
            source = row.source
            return partial(
                raw_row_count,
                client=client,
                raw_table_id=source,
            )
        raise ValueError(f"Unexpected API Call for row {row} and column {col}.")

    def format_result(self, result: object, row: AssetIndex, col: str) -> CellValue:
        if col == self.Columns.RowCount:
            if isinstance(result, int):
                if result == MAX_ROW_ITERATION_RUN_QUERY:
                    return f"≥{result:,}"
                else:
                    return result
            elif isinstance(result, str):
                return result
            return None
        elif isinstance(result, int | float | bool | str):
            return result
        elif col == self.Columns.DataSets:
            return result[0] if isinstance(result, list) and result and isinstance(result[0], str) else None
        elif col == self.Columns.Transformations:
            if isinstance(result, list) and len(result) > 0 and isinstance(result[0], Transformation):
                return f"{result[0].name} ({result[0].external_id})"
            return None
        elif col == self.Columns.ColumnCount:
            if isinstance(result, RawProfileResults):
                return result.column_count
            return None
        raise ValueError(f"unexpected result type {type(result)} for row {row!s} and column {col}.")

    def update_table(
        self,
        current_table: dict[tuple[AssetIndex, str], PendingCellValue],
        result: object,
        selected_row: AssetIndex,
        selected_col: str,
    ) -> dict[tuple[AssetIndex, str], PendingCellValue]:
        handlers: Mapping[
            str,
            Callable[
                [dict[tuple[AssetIndex, str], PendingCellValue], object, AssetIndex],
                dict[tuple[AssetIndex, str], PendingCellValue],
            ],
        ] = {
            self.Columns.Count: self._update_count,
            self.Columns.DataSets: self._update_datasets,
            self.Columns.DataSetCount: self._update_dataset_count,
            self.Columns.Transformations: self._update_transformations,
            self.Columns.ColumnCount: self._update_column_count,
        }
        handler = handlers.get(selected_col)
        if handler:
            return handler(current_table, result, selected_row)
        return current_table

    def _update_count(
        self,
        current_table: dict[tuple[AssetIndex, str], PendingCellValue],
        result: object,
        selected_row: AssetIndex,
    ) -> dict[tuple[AssetIndex, str], PendingCellValue]:
        new_table: dict[tuple[AssetIndex, str], PendingCellValue] = {}
        for (row, col), value in current_table.items():
            if row == selected_row and col == self.Columns.DataSets:
                new_table[(row, col)] = WaitingAPICall
            else:
                new_table[(row, col)] = value
        return new_table

    def _update_datasets(
        self,
        current_table: dict[tuple[AssetIndex, str], PendingCellValue],
        result: object,
        selected_row: AssetIndex,
    ) -> dict[tuple[AssetIndex, str], PendingCellValue]:
        if not (isinstance(result, list) and len(result) > 0 and all(isinstance(item, str) for item in result)):
            return current_table
        new_table: dict[tuple[AssetIndex, str], PendingCellValue] = {}
        for (row, col), value in current_table.items():
            if row != selected_row:
                new_table[(row, col)] = value
                continue
            for data_set in result:
                new_index = AssetIndex(
                    aggregator=selected_row.aggregator, source=selected_row.source, data_set_external_id=data_set
                )
                if col == self.Columns.DataSetCount:
                    new_table[(new_index, col)] = WaitingAPICall
                elif col == self.Columns.DataSets:
                    new_table[(new_index, col)] = data_set
                else:
                    new_table[(new_index, col)] = value
        return new_table

    def _update_dataset_count(
        self,
        current_table: dict[tuple[AssetIndex, str], PendingCellValue],
        result: object,
        selected_row: AssetIndex,
    ) -> dict[tuple[AssetIndex, str], PendingCellValue]:
        if not isinstance(result, int):
            return current_table
        new_table: dict[tuple[AssetIndex, str], PendingCellValue] = {}
        for (row, col), value in current_table.items():
            if row != selected_row:
                new_table[(row, col)] = value
                continue
            if col == self.Columns.Transformations:
                new_table[(row, col)] = WaitingAPICall
            else:
                new_table[(row, col)] = value
        return new_table

    def _update_transformations(
        self,
        current_table: dict[tuple[AssetIndex, str], PendingCellValue],
        result: object,
        selected_row: AssetIndex,
    ) -> dict[tuple[AssetIndex, str], PendingCellValue]:
        if not (
            isinstance(result, list) and len(result) > 0 and all(isinstance(item, Transformation) for item in result)
        ):
            return current_table
        sources_by_transformation_id = {
            transformation.id: [
                s for s in get_transformation_sources(transformation.query or "") if isinstance(s, RawTable)
            ]
            for transformation in result
        }
        new_table: dict[tuple[AssetIndex, str], PendingCellValue] = {}
        for (row, col), value in current_table.items():
            if row != selected_row:
                new_table[(row, col)] = value
                continue
            for transformation in result:
                sources = sources_by_transformation_id[transformation.id]
                if not sources:
                    new_table[(row, col)] = value
                    continue
                for source in sources:
                    new_index = AssetIndex(
                        aggregator=selected_row.aggregator,
                        data_set_external_id=selected_row.data_set_external_id,
                        source=source,
                    )
                    if col == self.Columns.RawTable:
                        new_table[(new_index, col)] = str(source)
                    elif col == self.Columns.RowCount:
                        # This will be updated by the ColumnCount API call which equals the /profiler/raw endpoint.
                        # If the profiles is not complete, we set it to WaitingAPICall to get the
                        # row count from the transformation preview.
                        new_table[(new_index, col)] = None
                    elif col == self.Columns.ColumnCount:
                        new_table[(new_index, col)] = WaitingAPICall
                    elif col == self.Columns.Transformations:
                        new_table[(new_index, col)] = f"{transformation.name} ({transformation.external_id})"
                    else:
                        new_table[(new_index, col)] = value
        return new_table

    def _update_column_count(
        self,
        current_table: dict[tuple[AssetIndex, str], PendingCellValue],
        result: object,
        selected_row: AssetIndex,
    ) -> dict[tuple[AssetIndex, str], PendingCellValue]:
        if not isinstance(result, RawProfileResults):
            return current_table
        new_table: dict[tuple[AssetIndex, str], PendingCellValue] = {}
        for (row, col), value in current_table.items():
            if row != selected_row:
                new_table[(row, col)] = value
                continue
            is_complete = result.is_complete and result.row_count < self.profile_row_limit
            if col == self.Columns.RowCount:
                # If the profile is complete, we can use the row count directly.
                # If not we set it to WaitingAPICall to get the row count from the transformation preview.
                new_table[(row, col)] = result.row_count if is_complete else WaitingAPICall
            elif col == self.Columns.ColumnCount:
                new_table[(row, col)] = result.column_count if is_complete else f"≥{result.column_count:,}"
            else:
                new_table[(row, col)] = value
        return new_table


class ProfileAssetCentricCommand(ProfileCommand[str]):
    def __init__(
        self,
        output_spreadsheet: Path | None = None,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
    ) -> None:
        super().__init__(output_spreadsheet, print_warning, skip_tracking, silent)
        self.hierarchy: str | None = None
        self.table_title = "Asset Centric Profile"
        self.aggregators: dict[str, AssetCentricAggregator] = {}

    class Columns:
        Resource = "Resource"
        Count = "Count"
        MetadataKeyCount = "Metadata Key Count"
        LabelCount = "Label Count"
        Transformation = "Transformations"

    def asset_centric(
        self, client: ToolkitClient, hierarchy: str | None = None, select_all: bool = False, verbose: bool = False
    ) -> list[dict[str, CellValue]]:
        if hierarchy is None and not select_all:
            self.hierarchy = AssetInteractiveSelect(client, "profile").select_hierarchy(allow_empty=True)
            self._ask_store_file()
        else:
            self.hierarchy = hierarchy
        if self.hierarchy is not None:
            self.table_title = f"Asset Centric Profile: {self.hierarchy}"
        self.aggregators.update(
            {
                agg.display_name: agg
                for agg in [
                    AssetAggregator(client),
                    EventAggregator(client),
                    FileAggregator(client),
                    TimeSeriesAggregator(client),
                    SequenceAggregator(client),
                ]
            }
        )
        if self.hierarchy is None:
            # Relationship and Labels does not belong to a specific hierarchy
            self.aggregators.update(
                {
                    agg.display_name: agg
                    for agg in [
                        RelationshipAggregator(client),
                        LabelCountAggregator(client),
                    ]
                }
            )
        result = self.create_profile_table(client, sheet=self.table_title)
        if self.output_spreadsheet:
            for aggregator in self.aggregators.values():
                if isinstance(aggregator, AssetAggregator):
                    used_metadata_keys = aggregator.used_metadata_keys(hierarchy=self.hierarchy)
                    self._write_to_spreadsheet(
                        [{"Metadata Key": key, "Count": count} for key, count in used_metadata_keys],
                        ["Metadata Key", "Count"],
                        self.output_spreadsheet,
                        sheet=aggregator.display_name,
                    )
        return result

    def create_initial_table(self, client: ToolkitClient) -> dict[tuple[str, str], PendingCellValue]:
        table: dict[tuple[str, str], str | int | float | bool | None | WaitingAPICallClass] = {}
        for index, aggregator in self.aggregators.items():
            table[(index, self.Columns.Resource)] = aggregator.display_name
            table[(index, self.Columns.Count)] = WaitingAPICall
            # Metadata Key count is only valid if we aggregate for all resources or assets.
            # Events/Files/TimeSeries/Sequences do not have a rootId to filter on.
            if isinstance(aggregator, MetadataAggregator) and (
                isinstance(aggregator, AssetAggregator) or self.hierarchy is None
            ):
                table[(index, self.Columns.MetadataKeyCount)] = WaitingAPICall
            else:
                table[(index, self.Columns.MetadataKeyCount)] = None
            if isinstance(aggregator, LabelAggregator) and (
                isinstance(aggregator, AssetAggregator) or self.hierarchy is None
            ):
                table[(index, self.Columns.LabelCount)] = WaitingAPICall
            else:
                table[(index, self.Columns.LabelCount)] = None
            table[(index, self.Columns.Transformation)] = WaitingAPICall if self.hierarchy is None else None
        return table

    def create_api_callable(self, row: str, col: str, client: ToolkitClient) -> Callable:
        aggregator = self.aggregators[row]
        if col == self.Columns.Count:
            return partial(aggregator.count, hierarchy=self.hierarchy)
        elif col == self.Columns.MetadataKeyCount and isinstance(aggregator, MetadataAggregator):
            return aggregator.metadata_key_count
        elif col == self.Columns.LabelCount and isinstance(aggregator, LabelAggregator):
            return aggregator.label_count
        elif col == self.Columns.Transformation:
            return aggregator.transformation_count
        raise ValueError(f"Unknown column: {col} for row: {row}")


class ProfileTransformationCommand(ProfileCommand[str]):
    def __init__(
        self,
        output_spreadsheet: Path | None = None,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
    ) -> None:
        super().__init__(output_spreadsheet, print_warning, skip_tracking, silent)
        self.table_title = "Transformation Profile"
        self.destination_type: AssetCentricDestinationType | None = None

    class Columns:
        Transformation = "Transformation"
        Source = "Sources"
        DestinationColumns = "Destination Columns"
        Destination = "Destination"
        ConflictMode = "Conflict Mode"
        IsPaused = "Is Paused"

    def transformation(
        self, client: ToolkitClient, destination_type: str | None = None, verbose: bool = False
    ) -> list[dict[str, CellValue]]:
        self.destination_type = AssetCentricDestinationSelect.get(destination_type)
        self.table_title = f"Transformation Profile destination: {self.destination_type}"
        return self.create_profile_table(client, sheet=self.destination_type)

    def create_initial_table(self, client: ToolkitClient) -> dict[tuple[str, str], PendingCellValue]:
        if self.destination_type is None:
            raise ToolkitValueError("Destination type must be set before calling create_initial_table.")
        iterable: Iterable[Transformation] = client.transformations.list(
            destination_type=self.destination_type, limit=-1
        )
        if self.destination_type == "assets":
            iterable = itertools.chain(iterable, client.transformations(destination_type="asset_hierarchy", limit=-1))
        table: dict[tuple[str, str], PendingCellValue] = {}
        for transformation in iterable:
            sources: list[SQLTable] = []
            destination_columns: list[str] = []
            if transformation.query:
                parser = SQLParser(transformation.query, operation="Profile transformations")
                sources = parser.sources
                destination_columns = parser.destination_columns
            index = str(transformation.id)
            table[(index, self.Columns.Transformation)] = transformation.name or transformation.external_id or "Unknown"
            table[(index, self.Columns.Source)] = ", ".join(map(str, sources))
            table[(index, self.Columns.DestinationColumns)] = (
                ", ".join(destination_columns) or None if destination_columns else None
            )
            table[(index, self.Columns.Destination)] = (
                transformation.destination.type or "Unknown" if transformation.destination else "Unknown"
            )
            table[(index, self.Columns.ConflictMode)] = transformation.conflict_mode or "Unknown"
            table[(index, self.Columns.IsPaused)] = (
                str(transformation.schedule.is_paused) if transformation.schedule else "No schedule"
            )
        return table

    def create_api_callable(self, row: str, col: str, client: ToolkitClient) -> Callable:
        raise NotImplementedError(f"{type(self).__name__} does not support API calls for {col} in row {row}.")


@dataclass(frozen=True)
class RawProfileIndex:
    raw_table: RawTable
    transformation_id: int | None = None


class ProfileRawCommand(ProfileCommand[RawProfileIndex]):
    class Columns:
        RAW = "Raw"
        Rows = "Rows"
        Columns = "Columns"
        Transformation = "Transformation"
        Destination = "Destination"
        ConflictMode = "ConflictMode"

    max_profile_raw_count = 10_000  # The number of rows to profile to get the number of columns.
    # The actual limit is 256 MB of data.
    # Ref https://github.com/cognitedata/profiler-api/blob/main/src/main/scala/com/cognite/raw_profiler/Profile.scala#L37
    profile_timeout_seconds = 60 * 4  # Timeout for the profiling operation in seconds,
    # This is the same ase the run/query maximum timeout.

    is_dynamic_table = True

    def __init__(
        self,
        output_spreadsheet: Path | None = None,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
    ) -> None:
        super().__init__(output_spreadsheet, print_warning, skip_tracking, silent)
        self.table_title = "RAW Profile"
        self.destination_type: AssetCentricDestinationType | None = None
        self.client: ToolkitClient | None = None

    def raw(
        self,
        client: ToolkitClient,
        destination_type: str | None = None,
        verbose: bool = False,
    ) -> list[dict[str, CellValue]]:
        self.destination_type = AssetCentricDestinationSelect.get(destination_type)
        self.table_title = f"RAW Profile destination: {self.destination_type}"
        return self.create_profile_table(client, sheet=self.destination_type)

    def create_initial_table(self, client: ToolkitClient) -> dict[tuple[RawProfileIndex, str], PendingCellValue]:
        if self.destination_type is None:
            raise ToolkitValueError("Destination type must be set before calling create_initial_table.")
        table: dict[tuple[RawProfileIndex, str], PendingCellValue] = {}
        existing_tables = self._get_existing_tables(client)
        transformations_by_raw_table = self._get_transformations_by_raw_table(client, self.destination_type)
        for raw_id, transformations in transformations_by_raw_table.items():
            is_missing = raw_id not in existing_tables
            missing_str = " (missing)" if is_missing else ""
            for transformation in transformations:
                index = RawProfileIndex(raw_table=raw_id, transformation_id=transformation.id)
                table[(index, self.Columns.RAW)] = f"{raw_id!s}{missing_str}"
                if is_missing:
                    table[(index, self.Columns.Rows)] = "N/A"
                    table[(index, self.Columns.Columns)] = "N/A"
                else:
                    # First, we request the API to get the column count. This uses the /profiler/raw endpoint,
                    # and if that is a complete profile, it will update the row count as well. If it is not complete,
                    # the raw count will be obtained from the transformation/preview endpoint by setting the
                    # table[(index, self.Columns.Rows)] to WaitingAPICall.
                    table[(index, self.Columns.Rows)] = None
                    table[(index, self.Columns.Columns)] = WaitingAPICall
                table[(index, self.Columns.Transformation)] = f"{transformation.name} ({transformation.external_id})"
                table[(index, self.Columns.Destination)] = (
                    transformation.destination.type if transformation.destination else "Unknown"
                )
                table[(index, self.Columns.ConflictMode)] = transformation.conflict_mode
        return table

    def create_api_callable(self, row: RawProfileIndex, col: str, client: ToolkitClient) -> Callable:
        if col == self.Columns.Columns:
            return partial(
                client.raw.profile,
                database=row.raw_table.db_name,
                table=row.raw_table.table_name,
                limit=self.max_profile_raw_count,
                timeout_seconds=self.profile_timeout_seconds,
            )
        elif col == self.Columns.Rows:
            return partial(
                raw_row_count,
                client=client,
                raw_table_id=row.raw_table,
            )
        raise ValueError(f"There are no API calls for {row} in column {col}.")

    def format_result(self, result: object, row: RawProfileIndex, col: str) -> CellValue:
        if col == self.Columns.Rows:
            if isinstance(result, int):
                if result == MAX_ROW_ITERATION_RUN_QUERY:
                    return f"≥{result:,}"
                else:
                    return result
            elif isinstance(result, str):
                return result
            return None
        if isinstance(result, int | float | bool | str) or result is None:
            return result
        elif isinstance(result, RawProfileResults) and col == self.Columns.Columns:
            return result.column_count
        raise ValueError(f"Unknown result type: {type(result)} for {row!s} in column {col}.")

    def update_table(
        self,
        current_table: dict[tuple[RawProfileIndex, str], PendingCellValue],
        result: object,
        selected_row: RawProfileIndex,
        selected_col: str,
    ) -> dict[tuple[RawProfileIndex, str], PendingCellValue]:
        if not isinstance(result, RawProfileResults) or selected_col != self.Columns.Columns:
            return current_table
        is_complete = result.is_complete and result.row_count < self.max_profile_raw_count
        new_table: dict[tuple[RawProfileIndex, str], PendingCellValue] = {}
        for (row, col), value in current_table.items():
            if row == selected_row and col == self.Columns.Rows:
                new_table[(row, col)] = result.row_count if is_complete else WaitingAPICall
            elif row == selected_row and col == self.Columns.Columns:
                new_table[(row, col)] = result.column_count if is_complete else f"≥{result.column_count:,}"
            else:
                new_table[(row, col)] = value
        return new_table

    @classmethod
    def _get_existing_tables(cls, client: ToolkitClient) -> set[RawTable]:
        existing_tables: set[RawTable] = set()
        databases = client.raw.databases.list(limit=-1)
        for database in databases:
            if database.name is None:
                continue
            tables = client.raw.tables.list(db_name=database.name, limit=-1)
            for table in tables:
                if table.name is None:
                    continue
                existing_tables.add(RawTable(db_name=database.name, table_name=table.name))
        return existing_tables

    @classmethod
    def _get_transformations_by_raw_table(
        cls, client: ToolkitClient, destination_type: str
    ) -> dict[RawTable, list[Transformation]]:
        transformations = client.transformations.list(destination_type=destination_type)
        if destination_type == "assets":
            transformations.extend(client.transformations.list(destination_type="asset_hierarchy"))
        transformations_by_raw_table: dict[RawTable, list[Transformation]] = defaultdict(list)
        for transformation in transformations:
            if transformation.query is None:
                # No query means no source table.
                continue
            sources = get_transformation_sources(transformation.query)
            for source in sources:
                if isinstance(source, RawTable):
                    transformations_by_raw_table[source].append(transformation)
        return transformations_by_raw_table



from pathlib import Path

import pytest
import yaml
from cognite.client.data_classes import (
    DataSet,
    ExtractionPipeline,
    ExtractionPipelineConfig,
    ExtractionPipelineConfigWrite,
    ExtractionPipelineWrite,
)
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import DumpResourceCommand
from cognite_toolkit._cdf_tk.commands.dump_resource import ExtractionPipelineFinder
from cognite_toolkit._cdf_tk.cruds import (
    ExtractionPipelineConfigCRUD,
    ExtractionPipelineCRUD,
)


@pytest.fixture(scope="session")
def deployed_extraction_pipeline(toolkit_client: ToolkitClient, toolkit_dataset: DataSet) -> ExtractionPipeline:
    pipeline = ExtractionPipelineWrite(
        external_id="toolkit_test_extraction_pipeline",
        name="Toolkit Test ExtractionPipeline",
        data_set_id=toolkit_dataset.id,
        description="This is used in integration tests of the toolkit.",
        created_by="Cognite Toolkit",
    )
    existing = toolkit_client.extraction_pipelines.retrieve(external_id=pipeline.external_id)
    if existing:
        return existing
    return toolkit_client.extraction_pipelines.create(pipeline)


@pytest.fixture(scope="session")
def deployed_extraction_pipeline_config(
    toolkit_client: ToolkitClient, deployed_extraction_pipeline: ExtractionPipeline
) -> ExtractionPipelineConfig:
    config = ExtractionPipelineConfigWrite(
        external_id=deployed_extraction_pipeline.external_id,
        config=yaml.safe_dump({"some": "config"}),
        description="This is used in integration tests of the toolkit.",
    )
    try:
        return toolkit_client.extraction_pipelines.config.retrieve(external_id=deployed_extraction_pipeline.external_id)
    except CogniteAPIError as e:
        if e.code == 404:
            # If the config does not exist, create it
            return toolkit_client.extraction_pipelines.config.create(config)
        raise e


class TestDumpExtractionPipeline:
    @pytest.mark.usefixtures(
        "deployed_extraction_pipeline_config",
    )
    def test_dump_extraction_pipeline_with_config(
        self,
        deployed_extraction_pipeline: ExtractionPipeline,
        toolkit_client: ToolkitClient,
        tmp_path: Path,
    ) -> None:
        cmd = DumpResourceCommand(silent=True)
        cmd.dump_to_yamls(
            ExtractionPipelineFinder(toolkit_client, (deployed_extraction_pipeline.external_id,)),
            output_dir=tmp_path,
            clean=False,
            verbose=False,
        )

        folder = tmp_path / ExtractionPipelineCRUD.folder_name
        assert folder.exists()
        assert sum(1 for _ in folder.glob(f"*{ExtractionPipelineCRUD.kind}.yaml")) == 1
        assert sum(1 for _ in folder.glob(f"*{ExtractionPipelineConfigCRUD.kind}.yaml")) == 1



from collections.abc import Sequence
from typing import Any, cast

from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.api_classes import PagedResponse, QueryResponse
from cognite_toolkit._cdf_tk.client.data_classes.infield import DataExplorationConfig, InfieldLocationConfig
from cognite_toolkit._cdf_tk.client.data_classes.instance_api import (
    InstanceResponseItem,
    InstanceResult,
    NodeIdentifier,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, ItemsRequest, SimpleBodyRequest


class InfieldConfigAPI:
    ENDPOINT = "/models/instances"
    LOCATION_REF = "locationConfig"
    EXPLORATION_REF = "explorerConfig"
    # We know that this key exists and it has alias set.
    DATA_EXPLORATION_PROP_ID = cast(str, InfieldLocationConfig.model_fields["data_exploration_config"].alias)

    def __init__(self, http_client: HTTPClient, console: Console) -> None:
        self._http_client = http_client
        self._console = console
        self._config = http_client.config

    def apply(self, items: Sequence[InfieldLocationConfig]) -> list[InstanceResult]:
        if len(items) > 500:
            raise ValueError("Cannot apply more than 500 InfieldLocationConfig items at once.")

        request_items = (
            [item.as_request_item()]
            if item.data_exploration_config is None
            else [item.as_request_item(), item.data_exploration_config.as_request_item()]
            for item in items
        )
        responses = self._http_client.request_with_retries(
            ItemsRequest(
                endpoint_url=self._config.create_api_url(self.ENDPOINT),
                method="POST",
                items=[item for sublist in request_items for item in sublist],
            )
        )
        responses.raise_for_status()
        return PagedResponse[InstanceResult].model_validate(responses.get_first_body()).items

    def retrieve(self, items: Sequence[NodeIdentifier]) -> list[InfieldLocationConfig]:
        if len(items) > 100:
            raise ValueError("Cannot retrieve more than 100 InfieldLocationConfig items at once.")
        if not items:
            return []
        responses = self._http_client.request_with_retries(
            SimpleBodyRequest(
                # We use the query endpoint to be able to retrieve linked DataExplorationConfig items
                endpoint_url=self._config.create_api_url(f"{self.ENDPOINT}/query"),
                method="POST",
                body_content=self._retrieve_query(items),
            )
        )
        responses.raise_for_status()
        parsed_response = QueryResponse[InstanceResponseItem].model_validate(responses.get_first_body())
        return self._parse_retrieve_response(parsed_response)

    def delete(self, items: Sequence[InfieldLocationConfig]) -> list[NodeIdentifier]:
        if len(items) > 500:
            raise ValueError("Cannot delete more than 500 InfieldLocationConfig items at once.")

        identifiers = (
            [item.as_id()]
            if item.data_exploration_config is None
            else [item.as_id(), item.data_exploration_config.as_id()]
            for item in items
        )
        responses = self._http_client.request_with_retries(
            ItemsRequest(
                endpoint_url=self._config.create_api_url(f"{self.ENDPOINT}/delete"),
                method="POST",
                items=[identifier for sublist in identifiers for identifier in sublist],
            )
        )
        responses.raise_for_status()
        return PagedResponse[NodeIdentifier].model_validate(responses.get_first_body()).items

    @classmethod
    def _retrieve_query(cls, items: Sequence[NodeIdentifier]) -> dict[str, Any]:
        return {
            "with": {
                cls.LOCATION_REF: {
                    "limit": len(items),
                    "nodes": {
                        "filter": {
                            "instanceReferences": [
                                {"space": item.space, "externalId": item.external_id} for item in items
                            ]
                        },
                    },
                },
                cls.EXPLORATION_REF: {
                    "nodes": {
                        "from": "locationConfig",
                        "direction": "outwards",
                        "through": {
                            "source": InfieldLocationConfig.VIEW_ID.dump(),
                            "identifier": cls.DATA_EXPLORATION_PROP_ID,
                        },
                    }
                },
            },
            "select": {
                cls.LOCATION_REF: {
                    "sources": [{"source": InfieldLocationConfig.VIEW_ID.dump(), "properties": ["*"]}],
                },
                cls.EXPLORATION_REF: {
                    "sources": [{"source": DataExplorationConfig.VIEW_ID.dump(), "properties": ["*"]}],
                },
            },
        }

    def _parse_retrieve_response(
        self, parsed_response: QueryResponse[InstanceResponseItem]
    ) -> list[InfieldLocationConfig]:
        data_exploration_results = (
            DataExplorationConfig.model_validate(
                item.get_properties_for_source(DataExplorationConfig.VIEW_ID, include_identifier=True)
            )
            for item in parsed_response.items[self.EXPLORATION_REF]
        )
        data_exploration_config_map = {(dec.space, dec.external_id): dec for dec in data_exploration_results}
        result: list[InfieldLocationConfig] = []
        for item in parsed_response.items[self.LOCATION_REF]:
            properties = item.get_properties_for_source(InfieldLocationConfig.VIEW_ID, include_identifier=True)
            data_exploration = properties.pop(self.DATA_EXPLORATION_PROP_ID, None)
            if isinstance(data_exploration, dict):
                space = data_exploration["space"]
                external_id = data_exploration["externalId"]
                if (space, external_id) not in data_exploration_config_map:
                    HighSeverityWarning(
                        f"{self.DATA_EXPLORATION_PROP_ID} with space '{space}' and externalId '{external_id}' referenced in InfieldLocationConfig '{properties['externalId']}' was not found in the retrieved results."
                    ).print_warning(console=self._console)
                else:
                    # Pydantic allow already validated models to be assigned to fields
                    properties[self.DATA_EXPLORATION_PROP_ID] = data_exploration_config_map[(space, external_id)]  # type: ignore[assignment,index]
            else:
                HighSeverityWarning(
                    f"InfieldLocationConfig '{properties['externalId']}' is missing a valid {self.DATA_EXPLORATION_PROP_ID} reference."
                ).print_warning(console=self._console)
            result.append(InfieldLocationConfig.model_validate(properties))
        return result


class InfieldAPI:
    def __init__(self, http_client: HTTPClient, console: Console) -> None:
        self._http_client = http_client
        self.config = InfieldConfigAPI(http_client, console)



import os
from collections.abc import Hashable
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from _pytest.monkeypatch import MonkeyPatch
from cognite.client.data_classes import ExtractionPipelineConfig, ExtractionPipelineConfigWrite
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.commands import CleanCommand
from cognite_toolkit._cdf_tk.cruds import (
    DataSetsCRUD,
    ExtractionPipelineConfigCRUD,
    ExtractionPipelineCRUD,
    RawDatabaseCRUD,
    RawTableCRUD,
    ResourceCRUD,
    ResourceWorker,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestExtractionPipelineDependencies:
    _yaml = """
        externalId: 'ep_src_asset_hamburg_sap'
        name: 'Hamburg SAP'
        dataSetId: 12345
    """

    config_yaml = """
        externalId: 'ep_src_asset'
        description: 'DB extractor config reading data from Springfield SAP'
    """

    def test_load_extraction_pipeline_upsert_update_one(
        self, toolkit_client_approval: ApprovalToolkitClient, monkeypatch: MonkeyPatch
    ) -> None:
        toolkit_client_approval.append(
            ExtractionPipelineConfig,
            ExtractionPipelineConfig(
                external_id="ep_src_asset",
                description="DB extractor config reading data from Springfield SAP",
                config="\n    logger: \n        {level: WARN}",
            ),
        )

        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = self.config_yaml

        loader = ExtractionPipelineConfigCRUD.create_loader(toolkit_client_approval.mock_client)
        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([local_file])
        assert {
            "create": len(resources.to_create),
            "changed": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "changed": 1, "delete": 0, "unchanged": 0}

    def test_load_extraction_pipeline_delete_one(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        env_vars_with_client: EnvironmentVariables,
        monkeypatch: MonkeyPatch,
    ) -> None:
        toolkit_client_approval.append(
            ExtractionPipelineConfig,
            ExtractionPipelineConfig(
                external_id="ep_src_asset",
                description="DB extractor config reading data from Springfield SAP",
                config="\n    logger: \n        {level: WARN}",
            ),
        )

        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = self.config_yaml
        local_file.stem = "ep_src_asset"

        cmd = CleanCommand(print_warning=False)
        loader = ExtractionPipelineConfigCRUD.create_loader(env_vars_with_client.get_client())
        with patch.object(ExtractionPipelineConfigCRUD, "find_files", return_value=[local_file]):
            res = cmd.clean_resources(loader, env_vars_with_client, [], dry_run=True, drop=True)
            assert res is not None
            assert res.deleted == 1


class TestExtractionPipelineLoader:
    @pytest.mark.parametrize(
        "item, expected",
        [
            pytest.param(
                {
                    "dataSetExternalId": "ds_my_dataset",
                    "rawTables": [
                        {"dbName": "my_db", "tableName": "my_table"},
                        {"dbName": "my_db", "tableName": "my_table2"},
                    ],
                },
                [
                    (DataSetsCRUD, "ds_my_dataset"),
                    (RawDatabaseCRUD, RawDatabase("my_db")),
                    (RawTableCRUD, RawTable("my_db", "my_table")),
                    (RawTableCRUD, RawTable("my_db", "my_table2")),
                ],
                id="Extraction pipeline to Table",
            ),
        ],
    )
    def test_get_dependent_items(self, item: dict, expected: list[tuple[type[ResourceCRUD], Hashable]]) -> None:
        actual = ExtractionPipelineCRUD.get_dependent_items(item)

        assert list(actual) == expected

    @patch.dict(
        os.environ,
        {
            "INGESTION_CLIENT_ID": "this-is-the-ingestion-client-id",
            "INGESTION_CLIENT_SECRET": "this-is-the-ingestion-client-secret",
            "NON-SECRET": "this-is-not-a-secret",
        },
    )
    def test_omit_environment_variables(
        self, env_vars_with_client: EnvironmentVariables, monkeypatch: MonkeyPatch
    ) -> None:
        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = """
            - externalId: 'ep_src_asset'
              name: 'Hamburg SAP'
              config: 'secret: ${INGESTION_CLIENT_SECRET}'
            - externalId: 'ep_src_asset_2'
              name: '${NON-SECRET}'
              config: 'secret: ${INGESTION_CLIENT_SECRET}'
        """
        local_file.stem = "ep_src_asset"

        loader = ExtractionPipelineConfigCRUD.create_loader(env_vars_with_client.get_client())
        res = loader.load_resource_file(filepath=local_file, environment_variables=env_vars_with_client.dump())
        # Assert that env vars are skipped for this loader
        assert res[0]["config"] == "secret: ${INGESTION_CLIENT_SECRET}"
        assert res[1]["name"] == "this-is-not-a-secret"


class TestExtractionPipelineConfigCRUD:
    def test_load_resource_no_warning_on_keyvault(self) -> None:
        resource = {
            "externalId": "ep_src_asset",
            "config": """azure-keyvault:
  authentication-method: client-secret
  keyvault-name: CogniteKeyVault
  tenant-id: ${AZ_ENTRA_TENANT_ID}
  client-id: ${AZ_SERVICE_PRINCIPLE_APPLICATION_ID}
  secret: ${AZ_SERVICE_PRINCIPLE_CLIENT_SECRET}
  password: !keyvault value-secret-name
databases:
-   connection-string:  !keyvault value-secret-name
    name: my_db
    type: odbc""",
        }
        console = MagicMock(spec=Console)
        print_mock = MagicMock()
        console.print = print_mock
        crud = ExtractionPipelineConfigCRUD(MagicMock(spec=ToolkitClient), None, console=console)

        loaded = crud.load_resource(resource)

        assert isinstance(loaded, ExtractionPipelineConfigWrite)
        # No warning should be printed
        print_mock.assert_not_called()

    def test_load_resource_invalid_yaml_warning(self) -> None:
        resource = {
            "externalId": "ep_src_asset",
            "config": "invalid-yaml: [unclosed_list",
        }
        console = MagicMock(spec=Console)
        print_mock = MagicMock()
        console.print = print_mock
        crud = ExtractionPipelineConfigCRUD(MagicMock(spec=ToolkitClient), None, console=console)
        loaded = crud.load_resource(resource)

        assert isinstance(loaded, ExtractionPipelineConfigWrite)
        print_mock.assert_called_once()
        args, _ = print_mock.call_args
        _, message = args
        assert "ep_src_asset" in message

    def test_load_resource_yaml_array(self) -> None:
        resource = {
            "externalId": "ep_src_asset",
            "config": "- item1: value1",
        }
        console = MagicMock(spec=Console)
        print_mock = MagicMock()
        console.print = print_mock
        crud = ExtractionPipelineConfigCRUD(MagicMock(spec=ToolkitClient), None, console=console)
        loaded = crud.load_resource(resource)

        assert isinstance(loaded, ExtractionPipelineConfigWrite)
        print_mock.assert_called_once()
        args, _ = print_mock.call_args
        _, message = args
        assert "ep_src_asset" in message
        assert "a valid YAML mapping" in message



from collections.abc import Hashable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes import data_modeling as dm

from cognite_toolkit._cdf_tk.cruds import ContainerCRUD, ResourceCRUD, ResourceWorker, SpaceCRUD, ViewCRUD
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestViewLoader:
    def test_unchanged_view_int_version(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        loader = ViewCRUD.create_loader(toolkit_client_approval.mock_client)
        raw_file = """- space: sp_space
  externalId: my_view
  version: 1"""
        file = MagicMock(spec=Path)
        file.read_text.return_value = raw_file
        cdf_view = dm.View(
            space="sp_space",
            external_id="my_view",
            version="1",
            last_updated_time=1,
            created_time=1,
            description=None,
            name=None,
            filter=None,
            implements=None,
            writable=True,
            used_for="node",
            is_global=False,
            properties={},
        )

        toolkit_client_approval.append(dm.View, [cdf_view])

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([file])
        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}

    @pytest.mark.parametrize(
        "item, expected",
        [
            pytest.param(
                {
                    "space": "sp_my_space",
                    "properties": {
                        "name": {
                            "container": {
                                "type": "container",
                                "space": "my_container_space",
                                "externalId": "my_container",
                            }
                        }
                    },
                },
                [
                    (SpaceCRUD, "sp_my_space"),
                    (ContainerCRUD, dm.ContainerId(space="my_container_space", external_id="my_container")),
                ],
                id="View with one container property",
            ),
            pytest.param(
                {
                    "space": "sp_my_space",
                    "properties": {
                        "toEdge": {
                            "source": {
                                "type": "view",
                                "space": "my_view_space",
                                "externalId": "my_view",
                                "version": "1",
                            },
                            "edgeSource": {
                                "type": "view",
                                "space": "my_other_view_space",
                                "externalId": "my_edge_view",
                                "version": "42",
                            },
                        }
                    },
                },
                [
                    (SpaceCRUD, "sp_my_space"),
                    (ViewCRUD, dm.ViewId(space="my_view_space", external_id="my_view", version="1")),
                    (ViewCRUD, dm.ViewId(space="my_other_view_space", external_id="my_edge_view", version="42")),
                ],
                id="View with one container property",
            ),
        ],
    )
    def test_get_dependent_items(self, item: dict, expected: list[tuple[type[ResourceCRUD], Hashable]]) -> None:
        actual = ViewCRUD.get_dependent_items(item)

        assert list(actual) == expected

    @pytest.mark.parametrize(
        "view_id,expected_readonly_props",
        [
            pytest.param(
                dm.ViewId("cdf_cdm", "CogniteAsset", "v1"),
                {"pathLastUpdatedTime", "path", "root"},
                id="CogniteAsset_has_readonly_properties",
            ),
            pytest.param(
                dm.ViewId("cdf_cdm", "CogniteFile", "v1"),
                {"isUploaded", "uploadedTime"},
                id="CogniteFile_has_readonly_properties",
            ),
            pytest.param(
                dm.ViewId("cdf_cdm", "CogniteSourceSystem", "v1"),
                set(),
                id="CogniteSourceSystem_no_readonly_properties",
            ),
        ],
    )
    def test_get_readonly_properties(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        cognite_core_no_3D: dm.DataModel,
        view_id: dm.ViewId,
        expected_readonly_props: set[str],
    ) -> None:
        """Test that get_readonly_properties identifies readonly properties from containers."""
        loader = ViewCRUD.create_loader(toolkit_client_approval.mock_client)
        toolkit_client_approval.append(dm.View, cognite_core_no_3D.views)

        readonly_props = loader.get_readonly_properties(view_id)
        assert readonly_props == expected_readonly_props



from collections.abc import Iterable
from pathlib import Path
from typing import Any

import pytest

from cognite_toolkit._cdf_tk.resource_classes.hosted_extractor_destination import HostedExtractorDestinationYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_hosted_extractor_destination_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "Pump"},
        {"Missing required field: 'externalId'", "Unused field: 'name'"},
        id="Missing required field: externalId",
    )
    yield pytest.param(
        {"externalId": "myDestination", "dataSetId": 123},
        {"Unused field: 'dataSetId'"},
        id="Unused field: dataSetId",
    )


class TestHostedExtractorDestinationYAML:
    @pytest.mark.parametrize("data", list(find_resources("Destination", resource_dir="hosted_extractors")))
    def test_load_valid_destination(self, data: dict[str, Any]) -> None:
        loaded = HostedExtractorDestinationYAML.model_validate(data)

        dumped = loaded.model_dump(exclude_unset=True, by_alias=True)
        if "credentials" in dumped:
            # Secret is not dumped as per design, so we add it back for comparison
            dumped["credentials"]["clientSecret"] = data["credentials"]["clientSecret"]
        assert dumped == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_hosted_extractor_destination_test_cases()))
    def test_invalid_destination_error_messages(self, data: dict[str, Any], expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, HostedExtractorDestinationYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



import pytest
from cognite.client.data_classes import (
    Asset,
)
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.utils.aggregators import (
    AssetAggregator,
    AssetCentricAggregator,
    EventAggregator,
    FileAggregator,
    SequenceAggregator,
    TimeSeriesAggregator,
)
from tests.test_integration.constants import (
    ASSET_COUNT,
    ASSET_DATASET,
    ASSET_TRANSFORMATION,
    EVENT_COUNT,
    EVENT_DATASET,
    EVENT_TRANSFORMATION,
    FILE_COUNT,
    FILE_DATASET,
    FILE_TRANSFORMATION,
    SEQUENCE_COUNT,
    SEQUENCE_DATASET,
    SEQUENCE_TRANSFORMATION,
    TIMESERIES_COUNT,
    TIMESERIES_DATASET,
    TIMESERIES_TRANSFORMATION,
)


class TestAggregators:
    aggregators: tuple[tuple[type[AssetCentricAggregator], str, str, int], ...] = (
        (AssetAggregator, ASSET_TRANSFORMATION, ASSET_DATASET, ASSET_COUNT),
        (TimeSeriesAggregator, TIMESERIES_TRANSFORMATION, TIMESERIES_DATASET, TIMESERIES_COUNT),
        (EventAggregator, EVENT_TRANSFORMATION, EVENT_DATASET, EVENT_COUNT),
        (FileAggregator, FILE_TRANSFORMATION, FILE_DATASET, FILE_COUNT),
        (SequenceAggregator, SEQUENCE_TRANSFORMATION, SEQUENCE_DATASET, SEQUENCE_COUNT),
    )

    @pytest.mark.usefixtures(
        "aggregator_assets", "aggregator_events", "aggregator_files", "aggregator_time_series", "aggregator_sequences"
    )
    @pytest.mark.parametrize(
        "aggregator_class, expected_transformation_external_id, expected_dataset_external_id, expected_count",
        aggregators,
    )
    def test_aggregations(
        self,
        toolkit_client: ToolkitClient,
        aggregator_class: type[AssetCentricAggregator],
        expected_transformation_external_id: str,
        expected_dataset_external_id: str,
        expected_count: int,
        aggregator_root_asset: Asset,
    ) -> None:
        root = aggregator_root_asset.external_id
        aggregator = aggregator_class(toolkit_client)

        try:
            actual_count = aggregator.count(root)
            used_data_sets = aggregator.used_data_sets(root)
            transformation_count = aggregator.transformation_count()
            used_transformations = aggregator.used_transformations(used_data_sets)
        except CogniteAPIError as e:
            if e.code == 500 and "Internal server error" in e.message:
                pytest.skip("Skipping test due to intermittent CDF 500 error.")
            raise e

        assert actual_count == expected_count
        assert used_data_sets == [expected_dataset_external_id]
        assert transformation_count >= 1  # We know at least one transformation is writing to the resource type.
        assert len(used_transformations) == 1
        assert used_transformations[0].external_id == expected_transformation_external_id



from abc import ABC
from typing import Any, Literal, TypeAlias

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    CogniteUpdate,
    ExternalIDTransformerMixin,
    PropertySpec,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from typing_extensions import Self

MapType: TypeAlias = Literal["WAYPOINTMAP", "THREEDMODEL", "TWODMAP", "POINTCLOUD"]


class RobotCapabilityCore(WriteableCogniteResource["RobotCapabilityWrite"], ABC):
    """Robot capabilities define what actions that robots can execute, including data capture (PTZ, PTZ-IR, 360)
    and behaviors (e.g., docking)

    Args:
        name: RobotCapability name.
        external_id: RobotCapability external id. Must be unique for the resource type.
        method: RobotCapability method. The method is used to call the right functionality on the robot.
        description: Description of RobotCapability. Textual description of the RobotCapability.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        method: str,
        description: str | None = None,
    ) -> None:
        self.name = name
        self.external_id = external_id
        self.method = method
        self.description = description


class RobotCapabilityWrite(RobotCapabilityCore):
    """Robot capabilities define what actions that robots can execute, including data capture (PTZ, PTZ-IR, 360)
    and behaviors (e.g., docking)

    Args:
        name: RobotCapability name.
        external_id: RobotCapability external id. Must be unique for the resource type.
        method: RobotCapability method. The method is used to call the right functionality on the robot.
        input_schema: Schema that defines what inputs are needed for the action. The input are values that
        configure the action, e.g pan, tilt and zoom values.
        data_handling_schema: Schema that defines how the data from a RobotCapability should be handled,
            including upload instructions.
        description: Description of RobotCapability. Textual description of the RobotCapability.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        method: str,
        input_schema: dict | None = None,
        data_handling_schema: dict | None = None,
        description: str | None = None,
    ) -> None:
        super().__init__(name, external_id, method, description)
        self.input_schema = input_schema
        self.data_handling_schema = data_handling_schema

    def as_write(self) -> Self:
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            method=resource["method"],
            input_schema=resource.get("inputSchema"),
            data_handling_schema=resource.get("dataHandlingSchema"),
            description=resource.get("description"),
        )


class RobotCapability(RobotCapabilityCore):
    """Robot capabilities define what actions that robots can execute, including data capture (PTZ, PTZ-IR, 360)
    and behaviors (e.g., docking)

    Args:
        name: RobotCapability name.
        external_id: RobotCapability external id. Must be unique for the resource type.
        method: RobotCapability method. The method is used to call the right functionality on the robot.
        input_schema: Schema that defines what inputs are needed for the action. The input are values that
        configure the action, e.g pan, tilt and zoom values.
        data_handling_schema: Schema that defines how the data from a RobotCapability should be handled,
            including upload instructions.
        description: Description of RobotCapability. Textual description of the RobotCapability.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        method: str,
        input_schema: dict,
        data_handling_schema: dict,
        description: str | None = None,
    ) -> None:
        super().__init__(name, external_id, method, description)
        self.input_schema = input_schema
        self.data_handling_schema = data_handling_schema

    def as_write(self) -> RobotCapabilityWrite:
        return RobotCapabilityWrite(
            name=self.name,
            external_id=self.external_id,
            method=self.method,
            input_schema=self.input_schema,
            data_handling_schema=self.data_handling_schema,
            description=self.description,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            method=resource["method"],
            input_schema=resource["inputSchema"],
            data_handling_schema=resource["dataHandlingSchema"],
            description=resource.get("description"),
        )


class RobotCapabilityWriteList(CogniteResourceList, ExternalIDTransformerMixin):
    _RESOURCE = RobotCapabilityWrite


class RobotCapabilityList(
    WriteableCogniteResourceList[RobotCapabilityWrite, RobotCapability], ExternalIDTransformerMixin
):
    _RESOURCE = RobotCapability

    def as_write(self) -> RobotCapabilityWriteList:
        return RobotCapabilityWriteList([capability.as_write() for capability in self])


class _RobotCapabilityUpdate(CogniteUpdate):
    """This is not fully implemented as the Toolkit only needs it for the
    _get_update_properties in the .update method of the RobotCapability class.

    All updates are done through the RobotCapabilityWrite class instead.
    """

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            # External ID is nullable, but is used in the upsert logic and thus cannot be nulled out.
            PropertySpec("external_id", is_nullable=False),
            PropertySpec("name"),
            PropertySpec("description", is_nullable=False),
            PropertySpec("method", is_nullable=False),
            PropertySpec("input_schema", is_nullable=False),
            PropertySpec("data_handling_schema", is_nullable=False),
        ]


RobotType: TypeAlias = Literal["SPOT", "ANYMAL", "DJI_DRONE", "TAUROB", "UWNKNOWN"]


class RobotCore(WriteableCogniteResource["RobotWrite"], ABC):
    """
    Robot contains information for a single robot, including capabilities, type, video streaming setup, etc.
    These fields are updated every time a robot registers with Robotics Services and do not require manual changes.

    Args:
        name: Robot name
        capabilities: List of external ids for the capabilities the robot can perform.
        robot_type: Type of robot.
        data_set_id: The id of the data set this asset belongs to.
        description: A brief description of the robot.
        metadata: Custom, application specific metadata. String key -> String value.
        location_external_id: External id of the location.

    """

    def __init__(
        self,
        name: str,
        capabilities: list[str],
        robot_type: RobotType,
        data_set_id: int,
        description: str | None = None,
        metadata: dict | None = None,
        location_external_id: str | None = None,
    ) -> None:
        self.name = name
        self.capabilities = capabilities
        self.robot_type = robot_type
        self.data_set_id = data_set_id
        self.description = description
        self.metadata = metadata
        self.location_external_id = location_external_id

    def as_write(self) -> "RobotWrite":
        return RobotWrite(
            name=self.name,
            capabilities=self.capabilities,
            robot_type=self.robot_type,
            data_set_id=self.data_set_id,
            description=self.description,
            metadata=self.metadata,
            location_external_id=self.location_external_id,
        )


class RobotWrite(RobotCore):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            capabilities=resource["capabilities"],
            robot_type=resource["robotType"],
            data_set_id=resource["dataSetId"],
            description=resource.get("description"),
            metadata=resource.get("metadata"),
            location_external_id=resource.get("locationExternalId"),
        )


class Robot(RobotCore):
    """
    Robot contains information for a single robot, including capabilities, type, video streaming setup, etc.
    These fields are updated every time a robot registers with Robotics Services and do not require manual changes.

    Args:
        name: Robot name
        capabilities: List of external ids for the capabilities the robot can perform.
        robot_type: Type of robot.
        data_set_id: The id of the data set this asset belongs to.
        created_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        updated_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        description: A brief description of the robot.
        metadata: Custom, application-specific metadata. String key -> String value.
        location_external_id: External id of the location.

    """

    def __init__(
        self,
        name: str,
        capabilities: list[str],
        robot_type: RobotType,
        data_set_id: int,
        created_time: int,
        updated_time: int,
        description: str | None = None,
        metadata: dict | None = None,
        location_external_id: str | None = None,
    ) -> None:
        super().__init__(name, capabilities, robot_type, data_set_id, description, metadata, location_external_id)
        self.created_time = created_time
        self.updated_time = updated_time

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            capabilities=resource["capabilities"],
            robot_type=resource["robotType"],
            data_set_id=resource["dataSetId"],
            created_time=resource["createdTime"],
            updated_time=resource["updatedTime"],
            description=resource.get("description"),
            metadata=resource.get("metadata"),
            location_external_id=resource.get("locationExternalId"),
        )


class RobotWriteList(CogniteResourceList):
    _RESOURCE = RobotWrite


class RobotList(WriteableCogniteResourceList[RobotWrite, Robot]):
    _RESOURCE = Robot

    def as_write(self) -> RobotWriteList:
        return RobotWriteList([robot.as_write() for robot in self])

    def get_robot_by_name(self, name: str) -> Robot:
        try:
            return next(robot for robot in self if robot.name == name)
        except StopIteration:
            raise ValueError(f"No robot with name {name} found in list")


class _RobotUpdate(CogniteUpdate):
    """This is not fully implemented as the Toolkit only needs it for the
    _get_update_properties in the .update method of the Robot class.

    All updates are done through the RobotWrite
    """

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("name", is_nullable=False),
            PropertySpec("description", is_nullable=False),
            PropertySpec("metadata", is_object=True, is_nullable=False),
            PropertySpec("robot_type", is_nullable=False),
            PropertySpec("location_external_id", is_nullable=False),
        ]


class DataPostProcessingCore(WriteableCogniteResource["DataPostProcessingWrite"], ABC):
    """Robot capabilities define what actions that robots can execute, including data capture (PTZ, PTZ-IR, 360)
    and behaviors (e.g., docking)

    Args:
        name: DataProcessing name.
        external_id: DataProcessing external id. Must be unique for the resource type.
        method: DataProcessing method. The method is used to call the right functionality on the robot.
        description: Description of DataProcessing. Textual description of the DataProcessing.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        method: str,
        description: str | None = None,
    ) -> None:
        self.name = name
        self.external_id = external_id
        self.method = method
        self.description = description


class DataPostProcessingWrite(DataPostProcessingCore):
    """DataPostprocessing define types of data processing on data captured by the robot.
    DataPostprocessing enables you to automatically process data captured by the robot.

    Args:
        name: DataProcessing name.
        external_id: DataProcessing external id. Must be unique for the resource type.
        method: DataProcessing method. The method is used to call the right functionality on the robot.
        input_schema: Schema that defines what inputs are needed for the data postprocessing. The input are values
            that configure the data postprocessing, e.g max and min values for a gauge.
        configure the action, e.g pan, tilt and zoom values.
        description: Description of DataProcessing. Textual description of the DataProcessing.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        method: str,
        input_schema: dict | None = None,
        description: str | None = None,
    ) -> None:
        super().__init__(name, external_id, method, description)
        self.input_schema = input_schema

    def as_write(self) -> Self:
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            method=resource["method"],
            input_schema=resource.get("inputSchema"),
            description=resource.get("description"),
        )


class DataPostProcessing(DataPostProcessingCore):
    """DataPostprocessing define types of data processing on data captured by the robot.
    DataPostprocessing enables you to automatically process data captured by the robot.

    Args:
        name: DataProcessing name.
        external_id: DataProcessing external id. Must be unique for the resource type.
        method: DataProcessing method. The method is used to call the right functionality on the robot.
        input_schema: Schema that defines what inputs are needed for the data postprocessing. The input are values
            that configure the data postprocessing, e.g max and min values for a gauge.
        configure the action, e.g pan, tilt and zoom values.
        description: Description of DataProcessing. Textual description of the DataProcessing.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        method: str,
        input_schema: dict,
        description: str | None = None,
    ) -> None:
        super().__init__(name, external_id, method, description)
        self.input_schema = input_schema

    def as_write(self) -> DataPostProcessingWrite:
        return DataPostProcessingWrite(
            name=self.name,
            external_id=self.external_id,
            method=self.method,
            input_schema=self.input_schema,
            description=self.description,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            method=resource["method"],
            input_schema=resource["inputSchema"],
            description=resource.get("description"),
        )


class DataPostProcessingWriteList(CogniteResourceList, ExternalIDTransformerMixin):
    _RESOURCE = DataPostProcessingWrite


class DataPostProcessingList(
    WriteableCogniteResourceList[DataPostProcessingWrite, DataPostProcessing], ExternalIDTransformerMixin
):
    _RESOURCE = DataPostProcessing

    def as_write(self) -> DataPostProcessingWriteList:
        return DataPostProcessingWriteList([capability.as_write() for capability in self])


class _DataProcessingUpdate(CogniteUpdate):
    """This is not fully implemented as the Toolkit only needs it for the
    _get_update_properties in the .update method of the DataProcessing class.

    All updates are done through the DataProcessingWrite class instead.
    """

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("name", is_nullable=False),
            PropertySpec("description", is_nullable=False),
            PropertySpec("method", is_nullable=False),
            PropertySpec("input_schema", is_nullable=False),
        ]


class LocationCore(WriteableCogniteResource["LocationWrite"], ABC):
    """Robot capabilities define what actions that robots can execute, including data capture (PTZ, PTZ-IR, 360)
    and behaviors (e.g., docking)

    Args:
        name: Location name.
        external_id: Location external id. Must be unique for the resource type.
        description: Description of Location. Textual description of the Location.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        description: str | None = None,
    ) -> None:
        self.name = name
        self.external_id = external_id
        self.description = description


class LocationWrite(LocationCore):
    """The Locations resource is used to specify the physical location of a robot. Robot missions are defined
    for a specific location. In addition, the location is used to group Missions and Map resources.

    Args:
        name: Location name.
        external_id: Location external id. Must be unique for the resource type.
        description: Description of Location. Textual description of the Location.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        description: str | None = None,
    ) -> None:
        super().__init__(name, external_id, description)

    def as_write(self) -> Self:
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            description=resource.get("description"),
        )


class Location(LocationCore):
    """DataPostprocessing define types of data processing on data captured by the robot.
    DataPostprocessing enables you to automatically process data captured by the robot.

    Args:
        name: Location name.
        external_id: Location external id. Must be unique for the resource type.
        created_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        updated_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        description: Description of Location. Textual description of the Location.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        created_time: int,
        updated_time: int,
        description: str | None = None,
    ) -> None:
        super().__init__(name, external_id, description)
        self.created_time = created_time
        self.updated_time = updated_time

    def as_write(self) -> LocationWrite:
        return LocationWrite(
            name=self.name,
            external_id=self.external_id,
            description=self.description,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            created_time=resource["createdTime"],
            updated_time=resource["updatedTime"],
            description=resource.get("description"),
        )


class LocationWriteList(CogniteResourceList, ExternalIDTransformerMixin):
    _RESOURCE = LocationWrite


class LocationList(WriteableCogniteResourceList[LocationWrite, Location], ExternalIDTransformerMixin):
    _RESOURCE = Location

    def as_write(self) -> LocationWriteList:
        return LocationWriteList([capability.as_write() for capability in self])


class _LocationUpdate(CogniteUpdate):
    """This is not fully implemented as the Toolkit only needs it for the
    _get_update_properties in the .update method of the Location class.

    All updates are done through the LocationWrite class instead.
    """

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("name", is_nullable=False),
            PropertySpec("description", is_nullable=False),
        ]


class Point3D(CogniteObject):
    """A point in 3D space."""

    def __init__(self, x: float, y: float, z: float) -> None:
        self.x = x
        self.y = y
        self.z = z

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            x=resource["x"],
            y=resource["y"],
            z=resource["z"],
        )


class Quaternion(Point3D):
    """A quaternion."""

    def __init__(self, x: float, y: float, z: float, w: float) -> None:
        super().__init__(x, y, z)
        self.w = w

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            x=resource["x"],
            y=resource["y"],
            z=resource["z"],
            w=resource["w"],
        )


class Transform(CogniteObject):
    """Transform of the parent frame to the current frame.

    Args:
        parent_frame_external_id: The external id of the parent frame.
        translation: Transform translation (Point3)
        orientation: Transform orientation as quaternion (Quaternion).

    """

    def __init__(
        self,
        parent_frame_external_id: str,
        translation: Point3D,
        orientation: Quaternion,
    ) -> None:
        self.parent_frame_external_id = parent_frame_external_id
        self.translation = translation
        self.orientation = orientation

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            parent_frame_external_id=resource["parentFrameExternalId"],
            translation=Point3D._load(resource["translation"]),
            orientation=Quaternion._load(resource["orientation"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "parentFrameExternalId" if camel_case else "parent_frame_external_id": self.parent_frame_external_id,
            "translation": self.translation.dump(camel_case),
            "orientation": self.orientation.dump(camel_case),
        }


class FrameCore(WriteableCogniteResource["FrameWrite"], ABC):
    """The frames resource represents coordinate frames, which are used to describe how maps are aligned with
    respect to each other. For example, frames are used to describe the relative position of a context map
    (e.g., a 3D model of a location) and a robot's navigation map. Frames are aligned with each other through
    transforms, which consist of a translation (in meters) and rotation (quaternion).

    Args:
        name: Frame name.
        external_id: Frame external id. Must be unique for the resource type.
        transform: Transform of the parent frame to the current frame.
    """

    def __init__(
        self,
        name: str,
        external_id: str,
        transform: Transform | None = None,
    ) -> None:
        self.name = name
        self.external_id = external_id
        self.transform = transform

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {
            "name": self.name,
            "externalId" if camel_case else "external_id": self.external_id,
        }
        if self.transform:
            output["transform"] = self.transform.dump(camel_case)
        return output


class FrameWrite(FrameCore):
    """The frames resource represents coordinate frames, which are used to describe how maps are aligned with
    respect to each other. For example, frames are used to describe the relative position of a context map
    (e.g., a 3D model of a location) and a robot's navigation map. Frames are aligned with each other through
    transforms, which consist of a translation (in meters) and rotation (quaternion).

    Args:
        name: Frame name.
        external_id: Frame external id. Must be unique for the resource type.
        transform: Transform of the parent frame to the current frame.
    """

    def __init__(
        self,
        name: str,
        external_id: str,
        transform: Transform | None = None,
    ) -> None:
        super().__init__(name, external_id, transform)

    def as_write(self) -> Self:
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            transform=Transform._load(resource["transform"], cognite_client) if resource.get("transform") else None,
        )


class Frame(FrameCore):
    """DataPostprocessing define types of data processing on data captured by the robot.
    DataPostprocessing enables you to automatically process data captured by the robot.

    Args:
        name: Frame name.
        external_id: Frame external id. Must be unique for the resource type.
        created_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        updated_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        created_time: int,
        updated_time: int,
        transform: Transform | None = None,
    ) -> None:
        super().__init__(name, external_id, transform)
        self.created_time = created_time
        self.updated_time = updated_time

    def as_write(self) -> FrameWrite:
        return FrameWrite(
            name=self.name,
            external_id=self.external_id,
            transform=self.transform,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            created_time=resource["createdTime"],
            updated_time=resource["updatedTime"],
            transform=Transform._load(resource["transform"]) if resource.get("transform") else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        output["createdTime" if camel_case else "created_time"] = self.created_time
        output["updatedTime" if camel_case else "updated_time"] = self.updated_time
        return output


class FrameWriteList(CogniteResourceList, ExternalIDTransformerMixin):
    _RESOURCE = FrameWrite


class FrameList(WriteableCogniteResourceList[FrameWrite, Frame], ExternalIDTransformerMixin):
    _RESOURCE = Frame

    def as_write(self) -> FrameWriteList:
        return FrameWriteList([capability.as_write() for capability in self])


class _FrameUpdate(CogniteUpdate):
    """This is not fully implemented as the Toolkit only needs it for the
    _get_update_properties in the .update method of the Frame class.

    All updates are done through the FrameWrite class instead.
    """

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("name", is_nullable=False),
            PropertySpec("transform", is_nullable=False),
        ]


class MapCore(WriteableCogniteResource["MapWrite"], ABC):
    """The map resource allows defining both context maps and robot maps of a specific location. A context map is a
    visual representation of a location, for example, a 3D model, a 2D floor plan, or a point cloud model.
    A robot map is a representation of where a robot is able to navigate. Maps need to be aligned with respect
    to each other using coordinate frames.

    Args:
        name: Map name.
        external_id: Map external id. Must be unique for the resource type.
        map_type: Map type
        description: Description of Map. Textual description of the Map.
        frame_external_id: External id of the map's reference frame.
        data: Map-specific data.
        location_external_id: External id of the location.
        scale: Uniform scaling factor, for example, for map unit conversion (centimeter to meter).

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        map_type: MapType,
        description: str | None = None,
        frame_external_id: str | None = None,
        data: dict | None = None,
        location_external_id: str | None = None,
        scale: float | None = None,
    ) -> None:
        self.name = name
        self.external_id = external_id
        self.map_type = map_type
        self.description = description
        self.frame_external_id = frame_external_id
        self.data = data
        self.location_external_id = location_external_id
        self.scale = scale


class MapWrite(MapCore):
    """The map resource allows defining both context maps and robot maps of a specific location. A context map is a
    visual representation of a location, for example, a 3D model, a 2D floor plan, or a point cloud model.
    A robot map is a representation of where a robot is able to navigate. Maps need to be aligned with respect
    to each other using coordinate frames.

    Args:
        name: Map name.
        external_id: Map external id. Must be unique for the resource type.
        map_type: Map type
        description: Description of Map. Textual description of the Map.
        frame_external_id: External id of the map's reference frame.
        data: Map-specific data.
        location_external_id: External id of the location.
        scale: Uniform scaling factor, for example, for map unit conversion (centimeter to meter).

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        map_type: MapType,
        description: str | None = None,
        frame_external_id: str | None = None,
        data: dict | None = None,
        location_external_id: str | None = None,
        scale: float | None = None,
    ) -> None:
        super().__init__(name, external_id, map_type, description, frame_external_id, data, location_external_id, scale)

    def as_write(self) -> Self:
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            map_type=resource["mapType"],
            description=resource.get("description"),
            frame_external_id=resource.get("frameExternalId"),
            data=resource.get("data"),
            location_external_id=resource.get("locationExternalId"),
            scale=resource.get("scale"),
        )


class Map(MapCore):
    """The map resource allows defining both context maps and robot maps of a specific location. A context map is a
    visual representation of a location, for example, a 3D model, a 2D floor plan, or a point cloud model.
    A robot map is a representation of where a robot is able to navigate. Maps need to be aligned with respect
    to each other using coordinate frames.

    Args:
        name: Map name.
        external_id: Map external id. Must be unique for the resource type.
        map_type: Map type
        created_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        updated_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        description: Description of Map. Textual description of the Map.
        frame_external_id: External id of the map's reference frame.
        data: Map-specific data.
        location_external_id: External id of the location.
        scale: Uniform scaling factor, for example, for map unit conversion (centimeter to meter).

    """

    def __init__(
        self,
        name: str,
        external_id: str,
        map_type: MapType,
        created_time: int,
        updated_time: int,
        description: str | None = None,
        frame_external_id: str | None = None,
        data: dict | None = None,
        location_external_id: str | None = None,
        scale: float | None = None,
    ) -> None:
        super().__init__(name, external_id, map_type, description, frame_external_id, data, location_external_id, scale)
        self.created_time = created_time
        self.updated_time = updated_time

    def as_write(self) -> MapWrite:
        return MapWrite(
            name=self.name,
            external_id=self.external_id,
            map_type=self.map_type,
            description=self.description,
            frame_external_id=self.frame_external_id,
            data=self.data,
            location_external_id=self.location_external_id,
            scale=self.scale,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            external_id=resource["externalId"],
            map_type=resource["mapType"],
            created_time=resource["createdTime"],
            updated_time=resource["updatedTime"],
            description=resource.get("description"),
            frame_external_id=resource.get("frameExternalId"),
            data=resource.get("data"),
            location_external_id=resource.get("locationExternalId"),
            scale=resource.get("scale"),
        )


class MapWriteList(CogniteResourceList, ExternalIDTransformerMixin):
    _RESOURCE = MapWrite


class MapList(WriteableCogniteResourceList[MapWrite, Map], ExternalIDTransformerMixin):
    _RESOURCE = Map

    def as_write(self) -> MapWriteList:
        return MapWriteList([capability.as_write() for capability in self])


class _MapUpdate(CogniteUpdate):
    """This is not fully implemented as the Toolkit only needs it for the
    _get_update_properties in the .update method of the Map class.

    All updates are done through the MapWrite class instead.
    """

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("name", is_nullable=False),
            PropertySpec("description", is_nullable=False),
            PropertySpec("frame_external_id", is_nullable=False),
            PropertySpec("data", is_nullable=False),
            PropertySpec("location_external_id", is_nullable=False),
            PropertySpec("scale", is_nullable=False),
        ]



from pydantic import Field

from .authentication import AuthenticationClientIdSecret
from .base import ToolkitResource


class FunctionScheduleYAML(ToolkitResource):
    function_external_id: str = Field(
        description="The external ID provided by the client for the function.",
        max_length=255,
    )
    name: str = Field(
        description="The name of the function schedule.",
        min_length=1,
        max_length=140,
    )
    cron_expression: str = Field(
        description="Cron expression.",
        min_length=1,
        max_length=1024,
    )
    description: str | None = Field(
        default=None,
        description="The description of the function schedule.",
        min_length=1,
        max_length=500,
    )
    data: dict[str, object] | None = Field(
        default=None,
        description="Input data to the function.",
    )
    authentication: AuthenticationClientIdSecret | None = Field(
        default=None, description="Credentials required for the authentication."
    )



import contextlib
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import pytest
from cognite.client.data_classes import (
    Asset,
    AssetWrite,
    DataSet,
    DataSetWrite,
    Event,
    EventWrite,
    ExtractionPipeline,
    ExtractionPipelineWrite,
    FileMetadata,
    FileMetadataWrite,
    LabelDefinition,
    LabelDefinitionWrite,
    Relationship,
    RelationshipWrite,
    Sequence,
    SequenceColumnWrite,
    SequenceWrite,
    ThreeDModel,
    ThreeDModelWrite,
    TimeSeries,
    TimeSeriesWrite,
    Transformation,
    TransformationWrite,
    Workflow,
    WorkflowUpsert,
)
from cognite.client.data_classes.data_modeling import NodeId, Space
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFileApply, CogniteTimeSeriesApply
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils import datetime_to_ms

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import PurgeCommand
from cognite_toolkit._cdf_tk.storageio.selectors import InstanceFileSelector
from tests.test_integration.constants import RUN_UNIQUE_ID


@pytest.fixture()
def file_ts_nodes(
    toolkit_client_with_pending_ids: ToolkitClient, toolkit_space: Space
) -> Iterable[tuple[tuple[NodeId, int], tuple[NodeId, int]]]:
    client = toolkit_client_with_pending_ids
    file = CogniteFileApply(
        space=toolkit_space.space,
        external_id=f"test_file_purge_with_unlink_{RUN_UNIQUE_ID}",
        name="Test File for Purge with Unlink",
        mime_type="text/plain",
    )
    ts = CogniteTimeSeriesApply(
        space=toolkit_space.space,
        external_id=f"test_ts_purge_with_unlink_{RUN_UNIQUE_ID}",
        name="Test TS for Purge with Unlink",
        is_step=False,
        time_series_type="numeric",
    )
    classic_file = FileMetadataWrite(
        name=file.name,
        external_id=file.external_id,
        mime_type=file.mime_type,
    )
    classic_ts = TimeSeriesWrite(
        external_id=ts.external_id,
        name=ts.name,
        is_step=ts.is_step,
        is_string=ts.time_series_type == "string",
    )
    file_id: int | None = None
    ts_id: int | None = None
    try:
        # Ensure clean state
        client.data_modeling.instances.delete([file.as_id(), ts.as_id()])
        client.files.delete(external_id=classic_file.external_id, ignore_unknown_ids=True)
        client.time_series.delete(external_id=classic_ts.external_id, ignore_unknown_ids=True)

        # Create timeseries and file
        created_file = client.files.upload_bytes(b"Sample file content", **classic_file.dump(camel_case=False))
        file_id = created_file.id
        created_ts = client.time_series.create(classic_ts)
        ts_id = created_ts.id
        client.time_series.data.insert(
            datapoints=[{"timestamp": datetime_to_ms(datetime(2020, 1, 1, 0, 0, 0)), "value": 1.0}],
            id=ts_id,
        )

        # Link them.
        client.files.set_pending_ids(file.as_id(), id=file_id)
        client.time_series.set_pending_ids(ts.as_id(), id=ts_id)

        # Create Nodes in CDM
        created = client.data_modeling.instances.apply([file, ts])
        assert len(created.nodes) == 2

        yield (file.as_id(), file_id), (ts.as_id(), ts_id)
    finally:
        client.data_modeling.instances.delete([file.as_id(), ts.as_id()])
        if file_id is not None:
            client.files.unlink_instance_ids(id=file_id)
            client.files.delete(id=file_id, ignore_unknown_ids=True)
        if ts_id is not None:
            client.time_series.unlink_instance_ids(id=ts_id)
            client.time_series.delete(id=ts_id, ignore_unknown_ids=True)


@dataclass
class PopulatedDataSet:
    dataset: DataSet
    asset: Asset
    event: Event
    sequence: Sequence
    timeseries: TimeSeries
    file: FileMetadata
    label: LabelDefinition
    relationships: Relationship
    three_d: ThreeDModel
    workflow: Workflow
    transformation: Transformation
    extraction_pipeline: ExtractionPipeline


@pytest.fixture()
def populated_dataset(toolkit_client: ToolkitClient) -> Iterable[PopulatedDataSet]:
    client = toolkit_client
    dataset = DataSetWrite(name="toolkit_test_purge_dataset", external_id="toolkit_test_purge_dataset")
    created = client.data_sets.retrieve(external_id=dataset.external_id)
    if not created:
        # DataSet cannot be deleted, so we create it only once and reuse it
        created = client.data_sets.create(dataset)

    asset = AssetWrite(
        name="Test Asset",
        external_id=f"test_asset_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
    )
    created_asset = client.assets.create(asset)

    event = EventWrite(
        external_id=f"test_event_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
    )
    created_event = client.events.create(event)

    sequence = SequenceWrite(
        external_id=f"test_sequence_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
        columns=[SequenceColumnWrite(external_id="col1", value_type="String")],
    )
    created_sequence = client.sequences.create(sequence)

    timeseries = TimeSeriesWrite(
        external_id=f"test_timeseries_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
    )
    created_timeseries = client.time_series.create(timeseries)

    file = FileMetadataWrite(
        name="Test File",
        external_id=f"test_file_{RUN_UNIQUE_ID}",
        mime_type="text/plain",
        data_set_id=created.id,
    )
    created_file, _ = client.files.create(file)

    label = LabelDefinitionWrite(
        name="Test Label",
        external_id=f"test_label_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
    )
    created_label = client.labels.create(label)

    relationship = RelationshipWrite(
        external_id=f"test_relationship_{RUN_UNIQUE_ID}",
        source_external_id=created_asset.external_id,
        target_external_id=created_event.external_id,
        source_type="asset",
        target_type="event",
        data_set_id=created.id,
    )
    created_relationship = client.relationships.create(relationship)

    three_d = ThreeDModelWrite(
        name=f"Test 3D Model {RUN_UNIQUE_ID}",
        data_set_id=created.id,
    )
    created_three_d = client.three_d.models.create(three_d)

    workflow = WorkflowUpsert(
        external_id=f"test_workflow_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
    )
    created_workflow = client.workflows.upsert(workflow)

    transformation = TransformationWrite(
        name="Test Transformation",
        external_id=f"test_transformation_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
        ignore_null_fields=True,
    )
    created_transformation = client.transformations.create(transformation)

    extraction_pipeline = ExtractionPipelineWrite(
        name="Test Extraction Pipeline",
        external_id=f"test_extraction_pipeline_{RUN_UNIQUE_ID}",
        data_set_id=created.id,
    )
    created_extraction_pipeline = client.extraction_pipelines.create(extraction_pipeline)

    yield PopulatedDataSet(
        dataset=created,
        asset=created_asset,
        event=created_event,
        sequence=created_sequence,
        timeseries=created_timeseries,
        file=created_file,
        label=created_label,
        relationships=created_relationship,
        three_d=created_three_d,
        workflow=created_workflow,
        transformation=created_transformation,
        extraction_pipeline=created_extraction_pipeline,
    )

    # Cleanup
    client.assets.delete(id=created_asset.id, ignore_unknown_ids=True)
    client.events.delete(id=created_event.id, ignore_unknown_ids=True)
    client.sequences.delete(id=created_sequence.id, ignore_unknown_ids=True)
    client.time_series.delete(id=created_timeseries.id, ignore_unknown_ids=True)
    client.files.delete(id=created_file.id, ignore_unknown_ids=True)
    client.labels.delete(external_id=created_label.external_id)
    client.relationships.delete(external_id=created_relationship.external_id, ignore_unknown_ids=True)
    with contextlib.suppress(CogniteAPIError):
        client.three_d.models.delete(id=created_three_d.id)
    client.workflows.delete(external_id=created_workflow.external_id, ignore_unknown_ids=True)
    client.transformations.delete(id=created_transformation.id, ignore_unknown_ids=True)
    with contextlib.suppress(CogniteNotFoundError):
        client.extraction_pipelines.delete(id=created_extraction_pipeline.id)


class TestPurge:
    def test_purge_instances_with_unlink(
        self,
        file_ts_nodes: tuple[tuple[NodeId, int], tuple[NodeId, int]],
        toolkit_client_with_pending_ids: ToolkitClient,
        tmp_path: Path,
    ) -> None:
        client = toolkit_client_with_pending_ids
        (file_node, file_id), (ts_node, ts_id) = file_ts_nodes

        csv_path = tmp_path / "test.csv"
        csv_path.write_text(
            f"""space,externalId,instanceType
{file_node.space},{file_node.external_id},node
{ts_node.space},{ts_node.external_id},node
""",
            encoding="utf-8",
        )

        purge = PurgeCommand(silent=True)

        results = purge.instances(
            client,
            InstanceFileSelector(datafile=csv_path, validate=True),
            dry_run=False,
            unlink=True,
            verbose=False,
            auto_yes=True,
        )
        assert results.deleted == 2

        results = client.data_modeling.instances.retrieve([file_node, ts_node])
        assert len(results.nodes) == 0, "Instances were not purged"

        classic_file = client.files.retrieve(id=file_id)
        assert classic_file is not None, "File was not unlinked"

        classic_ts = client.time_series.retrieve(id=ts_id)
        assert classic_ts is not None, "Time series was not unlinked"

    def test_purge_dataset_include_data(
        self, toolkit_client: ToolkitClient, populated_dataset: PopulatedDataSet
    ) -> None:
        client = toolkit_client
        populated = populated_dataset
        purge = PurgeCommand(silent=True)

        _ = purge.dataset(
            client,
            selected_data_set_external_id=populated.dataset.external_id,
            archive_dataset=False,
            include_data=True,
            include_configurations=False,
            dry_run=False,
            auto_yes=True,
            verbose=False,
        )
        # Data is deleted
        assert client.assets.retrieve(external_id=populated.asset.external_id) is None
        assert client.events.retrieve(external_id=populated.event.external_id) is None
        assert client.sequences.retrieve(external_id=populated.sequence.external_id) is None
        assert client.time_series.retrieve(external_id=populated.timeseries.external_id) is None
        assert client.files.retrieve(external_id=populated.file.external_id) is None
        # Labels are not deleted, they are still available on direct look-up.
        # However, they should not be listed under the dataset anymore.
        assert len(client.labels.list(data_set_external_ids=populated.dataset.external_id)) == 0
        relationships = client.relationships.list(source_external_ids=[populated.asset.external_id])
        assert len(relationships) == 0
        assert client.three_d.models.retrieve(id=populated.three_d.id) is None

        # Configurations are not deleted
        assert (
            client.workflows.retrieve(external_id=populated.workflow.external_id, ignore_unknown_ids=True) is not None
        )
        assert client.transformations.retrieve(external_id=populated.transformation.external_id) is not None
        assert client.extraction_pipelines.retrieve(external_id=populated.extraction_pipeline.external_id) is not None

    def test_purge_dataset_include_configurations(
        self, toolkit_client: ToolkitClient, populated_dataset: PopulatedDataSet
    ) -> None:
        client = toolkit_client
        populated = populated_dataset
        purge = PurgeCommand(silent=True)

        _ = purge.dataset(
            client,
            selected_data_set_external_id=populated.dataset.external_id,
            archive_dataset=False,
            include_data=False,
            include_configurations=True,
            dry_run=False,
            auto_yes=True,
            verbose=False,
        )
        # Data not deleted
        assert client.assets.retrieve(external_id=populated.asset.external_id) is not None
        assert client.events.retrieve(external_id=populated.event.external_id) is not None
        assert client.sequences.retrieve(external_id=populated.sequence.external_id) is not None
        assert client.time_series.retrieve(external_id=populated.timeseries.external_id) is not None
        assert client.files.retrieve(external_id=populated.file.external_id) is not None
        # Labels are not deleted, they are still available on direct look-up.
        # However, they should not be listed under the dataset anymore.
        assert len(client.labels.list(data_set_external_ids=populated.dataset.external_id)) >= 1
        relationships = client.relationships.list(source_external_ids=[populated.asset.external_id])
        assert len(relationships) == 1
        assert client.three_d.models.retrieve(id=populated.three_d.id) is not None
        # Configurations deleted
        assert client.workflows.retrieve(external_id=populated.workflow.external_id, ignore_unknown_ids=True) is None
        assert client.transformations.retrieve(external_id=populated.transformation.external_id) is None
        assert client.extraction_pipelines.retrieve(external_id=populated.extraction_pipeline.external_id) is None

    def test_purge_dataset_dry_run(self, toolkit_client: ToolkitClient, populated_dataset: PopulatedDataSet) -> None:
        client = toolkit_client
        populated = populated_dataset
        purge = PurgeCommand(silent=True)

        results = purge.dataset(
            client,
            selected_data_set_external_id=populated.dataset.external_id,
            archive_dataset=False,
            include_data=True,
            include_configurations=True,
            dry_run=True,
            auto_yes=True,
            verbose=False,
        )
        assert results.dry_run == 1

        # Data not deleted
        assert client.assets.retrieve(external_id=populated.asset.external_id) is not None
        assert client.events.retrieve(external_id=populated.event.external_id) is not None
        assert client.sequences.retrieve(external_id=populated.sequence.external_id) is not None
        assert client.time_series.retrieve(external_id=populated.timeseries.external_id) is not None
        assert client.files.retrieve(external_id=populated.file.external_id) is not None
        # Labels are not deleted, they are still available on direct look-up.
        # However, they should not be listed under the dataset anymore.
        assert len(client.labels.list(data_set_external_ids=populated.dataset.external_id)) >= 1
        relationships = client.relationships.list(source_external_ids=[populated.asset.external_id])
        assert len(relationships) == 1
        assert client.three_d.models.retrieve(id=populated.three_d.id) is not None
        # Configurations not deleted
        assert (
            client.workflows.retrieve(external_id=populated.workflow.external_id, ignore_unknown_ids=True) is not None
        )
        assert client.transformations.retrieve(external_id=populated.transformation.external_id) is not None
        assert client.extraction_pipelines.retrieve(external_id=populated.extraction_pipeline.external_id) is not None



import sys
from abc import ABC
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteResourceList, WriteableCogniteResourceList
from cognite.client.data_classes.data_modeling import DataModelId, ViewId
from cognite.client.data_classes.data_modeling.core import DataModelingSchemaResource

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class _GraphQLDataModelCore(DataModelingSchemaResource["GraphQLDataModelWrite"], ABC):
    def __init__(
        self, space: str, external_id: str, version: str, name: str | None = None, description: str | None = None
    ) -> None:
        super().__init__(space=space, external_id=external_id, name=name, description=description)
        self.version = version

    def as_id(self) -> DataModelId:
        return DataModelId(space=self.space, external_id=self.external_id, version=self.version)


class GraphQLDataModelWrite(_GraphQLDataModelCore):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: str,
        name: str | None = None,
        description: str | None = None,
        previous_version: str | None = None,
        dml: str | None = None,
        preserve_dml: bool | None = None,
    ) -> None:
        super().__init__(space=space, external_id=external_id, version=version, name=name, description=description)
        self.previous_version = previous_version
        self.dml = dml
        self.preserve_dml = preserve_dml

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            version=str(resource["version"]),
            name=resource.get("name"),
            description=resource.get("description"),
            previous_version=resource.get("previousVersion"),
            dml=resource.get("dml"),
            preserve_dml=resource.get("preserveDml"),
        )

    def as_write(self) -> Self:
        return self


class GraphQLDataModel(_GraphQLDataModelCore):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: str,
        is_global: bool,
        last_updated_time: int,
        created_time: int,
        description: str | None,
        name: str | None,
        views: list[ViewId] | None,
    ) -> None:
        super().__init__(space=space, external_id=external_id, version=version, name=name, description=description)
        self.is_global = is_global
        self.last_updated_time = last_updated_time
        self.created_time = created_time
        self.views = views

    def as_write(self) -> GraphQLDataModelWrite:
        return GraphQLDataModelWrite(
            space=self.space,
            external_id=self.external_id,
            version=self.version,
            name=self.name,
            description=self.description,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            version=str(resource["version"]),
            is_global=resource["isGlobal"],
            last_updated_time=resource["lastUpdatedTime"],
            created_time=resource["createdTime"],
            description=resource.get("description"),
            name=resource.get("name"),
            views=[
                ViewId(space=view["space"], external_id=view["externalId"], version=view.get("version"))
                for view in resource.get("views", [])
            ],
        )


class GraphQLDataModelWriteList(CogniteResourceList[GraphQLDataModelWrite]):
    _RESOURCE = GraphQLDataModelWrite

    def as_ids(self) -> list[DataModelId]:
        return [model.as_id() for model in self.data]


class GraphQLDataModelList(WriteableCogniteResourceList[GraphQLDataModelWrite, GraphQLDataModel]):
    _RESOURCE = GraphQLDataModel

    def as_write(self) -> GraphQLDataModelWriteList:
        return GraphQLDataModelWriteList([model.as_write() for model in self.data])

    def as_ids(self) -> list[DataModelId]:
        return [model.as_id() for model in self.data]



from abc import ABC
from dataclasses import dataclass
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResourceList,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from typing_extensions import Self


@dataclass(frozen=True)
class ViewId(CogniteObject):
    external_id: str
    space: str

    def __str__(self) -> str:
        return f"({self.external_id}, {self.space})"

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            space=resource["space"],
        )


@dataclass
class SearchConfigViewProperty(CogniteObject):
    property: str
    disabled: bool | None = None
    selected: bool | None = None
    hidden: bool | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            property=resource["property"],
            disabled=resource.get("disabled"),
            selected=resource.get("selected"),
            hidden=resource.get("hidden"),
        )


class SearchConfigCore(WriteableCogniteResource["SearchConfigWrite"], ABC):
    """
    Core model for a single Configuration.

    Args:
        view: The configuration for one specific view.
        id: A server-generated ID for the object.
        use_as_name: The name of property to use for the name column in the UI.
        use_as_description: The name of property to use for the description column in the UI.
        columns_layout: Array of column configurations per property.
        filter_layout: Array of filter configurations per property.
        properties_layout: Array of property configurations per property.
    """

    def __init__(
        self,
        view: ViewId,
        id: int | None = None,
        use_as_name: str | None = None,
        use_as_description: str | None = None,
        columns_layout: list[SearchConfigViewProperty] | None = None,
        filter_layout: list[SearchConfigViewProperty] | None = None,
        properties_layout: list[SearchConfigViewProperty] | None = None,
    ) -> None:
        self.view = view
        self.id = id
        self.use_as_name = use_as_name
        self.use_as_description = use_as_description
        self.columns_layout = columns_layout
        self.filter_layout = filter_layout
        self.properties_layout = properties_layout

    def as_write(self) -> "SearchConfigWrite":
        return SearchConfigWrite(
            view=self.view,
            id=self.id,
            use_as_name=self.use_as_name,
            use_as_description=self.use_as_description,
            columns_layout=self.columns_layout,
            filter_layout=self.filter_layout,
            properties_layout=self.properties_layout,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.columns_layout:
            output["columnsLayout" if camel_case else "columns_layout"] = [
                _data.dump(camel_case) for _data in self.columns_layout
            ]
        if self.filter_layout:
            output["filterLayout" if camel_case else "filter_layout"] = [
                _data.dump(camel_case) for _data in self.filter_layout
            ]
        if self.properties_layout:
            output["propertiesLayout" if camel_case else "properties_layout"] = [
                _data.dump(camel_case) for _data in self.properties_layout
            ]
        if self.view:
            output["view"] = self.view.dump(camel_case)
        return output


class SearchConfigWrite(SearchConfigCore):
    """
    SearchConfig write/requst format.

    Args:
        view: The configuration for one specific view.
        id: A server-generated ID for the object.
        use_as_name: The name of property to use for the name column in the UI.
        use_as_description: The name of property to use for the description column in the UI.
        columns_layout: Array of column configurations per property.
        filter_layout: Array of filter configurations per property.
        properties_layout: Array of property configurations per property.
    """

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            id=resource.get("id"),
            view=ViewId.load(resource["view"]),
            use_as_name=resource.get("useAsName"),
            use_as_description=resource.get("useAsDescription"),
            columns_layout=(
                [SearchConfigViewProperty.load(item) for item in items]
                if (items := resource.get("columnsLayout"))
                else None
            ),
            filter_layout=(
                [SearchConfigViewProperty.load(item) for item in items]
                if (items := resource.get("filterLayout"))
                else None
            ),
            properties_layout=(
                [SearchConfigViewProperty.load(item) for item in items]
                if (items := resource.get("propertiesLayout"))
                else None
            ),
        )


class SearchConfig(SearchConfigCore):
    """
    Response model for a single Configuration.

    Args:
        view: The configuration for one specific view.
        id: A server-generated ID for the object.
        created_time: The time when the search config was created.
        updated_time: The time when the search config was last updated.
        use_as_name: The name of property to use for the name column in the UI.
        use_as_description: The name of property to use for the description column in the UI.
        columns_layout: Array of column configurations per property.
        filter_layout: Array of filter configurations per property.
        properties_layout: Array of property configurations per property.
    """

    def __init__(
        self,
        view: ViewId,
        id: int,
        created_time: int,
        updated_time: int,
        use_as_name: str | None = None,
        use_as_description: str | None = None,
        columns_layout: list[SearchConfigViewProperty] | None = None,
        filter_layout: list[SearchConfigViewProperty] | None = None,
        properties_layout: list[SearchConfigViewProperty] | None = None,
    ) -> None:
        super().__init__(
            view,
            id,
            use_as_name,
            use_as_description,
            columns_layout,
            filter_layout,
            properties_layout,
        )
        self.created_time = created_time
        self.updated_time = updated_time

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            view=ViewId.load(resource["view"]),
            id=resource["id"],
            created_time=resource["createdTime"],
            updated_time=resource["lastUpdatedTime"],
            use_as_name=resource.get("useAsName"),
            use_as_description=resource.get("useAsDescription"),
            columns_layout=(
                [SearchConfigViewProperty.load(item) for item in items]
                if (items := resource.get("columnsLayout"))
                else None
            ),
            filter_layout=(
                [SearchConfigViewProperty.load(item) for item in items]
                if (items := resource.get("filterLayout"))
                else None
            ),
            properties_layout=(
                [SearchConfigViewProperty.load(item) for item in items]
                if (items := resource.get("propertiesLayout"))
                else None
            ),
        )


class SearchConfigWriteList(CogniteResourceList):
    _RESOURCE = SearchConfigWrite


class SearchConfigList(WriteableCogniteResourceList[SearchConfigWrite, SearchConfig]):
    _RESOURCE = SearchConfig

    def as_write(self) -> SearchConfigWriteList:
        return SearchConfigWriteList([searchConfig.as_write() for searchConfig in self])



import sys
from collections import UserList
from typing import Any, Literal

from cognite.client import CogniteClient

from cognite_toolkit._cdf_tk.constants import StreamTemplateName
from cognite_toolkit._cdf_tk.protocols import ResourceRequestListProtocol, ResourceResponseListProtocol

from .base import BaseModelObject, RequestResource, ResponseResource

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class StreamRequest(RequestResource):
    """Stream request resource class."""

    external_id: str
    settings: dict[Literal["template"], dict[Literal["name"], StreamTemplateName]]

    def as_id(self) -> str:
        return self.external_id


class StreamRequestList(UserList[StreamRequest], ResourceRequestListProtocol):
    """List of Stream request resources."""

    _RESOURCE = StreamRequest
    items: list[StreamRequest]

    def __init__(self, initlist: list[StreamRequest] | None = None, **_: Any) -> None:
        super().__init__(initlist or [])

    def dump(self, camel_case: bool = True) -> list[dict[str, Any]]:
        return [item.dump(camel_case) for item in self.items]

    @classmethod
    def load(cls, data: list[dict[str, Any]], cognite_client: CogniteClient | None = None) -> "StreamRequestList":
        items = [StreamRequest.model_validate(item) for item in data]
        return cls(items)


class LifecycleObject(BaseModelObject):
    """Lifecycle object."""

    hot_phase_duration: str | None = None
    data_deleted_after: str | None = None
    retained_after_soft_delete: str


class ResourceUsage(BaseModelObject):
    """Resource quota with provisioned and consumed values."""

    provisioned: int
    consumed: int | None = None


class LimitsObject(BaseModelObject):
    """Limits object."""

    max_records_total: ResourceUsage
    max_giga_bytes_total: ResourceUsage
    max_filtering_interval: str | None = None


class StreamSettings(BaseModelObject):
    """Stream settings object."""

    lifecycle: LifecycleObject
    limits: LimitsObject


class StreamResponse(ResponseResource["StreamRequest"]):
    """Stream response resource class."""

    external_id: str
    created_time: int
    created_from_template: StreamTemplateName | None = None
    type: Literal["Mutable", "Immutable"] | None = None
    settings: StreamSettings | None = None

    def as_request_resource(self) -> StreamRequest:
        template_name = self.created_from_template or "BasicArchive"
        return StreamRequest(
            external_id=self.external_id,
            settings={"template": {"name": template_name}},
        )

    def as_write(self) -> Self:
        return self


class StreamResponseList(UserList[StreamResponse], ResourceResponseListProtocol):
    """List of Stream response resources."""

    _RESOURCE = StreamResponse
    data: list[StreamResponse]

    def __init__(self, initlist: list[StreamResponse] | None = None, **_: Any) -> None:
        super().__init__(initlist or [])

    def dump(self, camel_case: bool = True) -> list[dict[str, Any]]:
        return [item.dump(camel_case) for item in self.data]

    @classmethod
    def load(cls, data: list[dict[str, Any]], cognite_client: CogniteClient | None = None) -> "StreamResponseList":
        items = [StreamResponse.model_validate(item) for item in data]
        return cls(items)

    def as_write(self) -> Self:
        return self
